# History: Apr 21 2011
#          Oct 08 2012 Add options
#          Nov 09 2013 Add options

crave_user_func <- function(delta, cov, cor, invcor, user_fnc) {

  # Convert vectors to matrices
  n <- length(delta)
  cov <- matrix(cov, nrow=n, ncol=n, byrow=TRUE)
  cor <- matrix(cor, nrow=n, ncol=n, byrow=TRUE)
  invcor <- matrix(invcor, nrow=n, ncol=n, byrow=TRUE)
  
  str <- paste(user_fnc[5], "(delta, cov, cor, invcor)", sep="") 
  ret <- eval(parse(text=str))
  ret
}

crave_print <- function(str, which, sval, dval, ival) {

  #str <- sub("\n", "", str, fixed=TRUE)
  if (!which) {
    cat(str)
  } else {
    x <- strsplit(str, "%", fixed=TRUE)[[1]]
    if (which == 1) {
      value <- sval[3]
    } else if (which == 2) {
      value <- dval
    } else {
      value = ival
    }
    str <- paste(x[1], value, substring(x[2], 2), sep="")
    cat(str)
  }

  0
}

crave <- function(geno_file, pheno_file, out_gene, out_snp, op=NULL) {

  op <- default.list(op, 
         c("notGroupedByChr", "max_maf", "max_miss_rate",
           "pheno_miss", "covars", "gene_file", "reflink.geneName",
           "nperm", "adapt_maxperm", "adapt_tol", "das_alpha",
           "threshold_alpha", "bm_n", "print", "tests", "no_gene_file",
           "rover_t", "vs", "update_cor", "store_perms", "gene_kb"
         ), 
         list(FALSE, 0.5, 0.8,
              '.', 0, "hg19", "reflink.geneName",
              500, 1000000, 0.1, 0.05,
              0.05, 10, 0, "MDF", FALSE,
              0.2, 0.005, 0.3, 0, 10)    
         )

  # Get the default gene_file if needed
  gene_file <- op[["gene_file", exact=TRUE]]
  if (is.null(gene_file)) {
    gene_file <- system.file("sampleData", "chr_1.txt.gz", package="CRAVE")
    gene_file <- dirname(gene_file)
  } else if (gene_file %in% c("hg19")) {
    gene_file <- system.file("sampleData", "chr_1.txt.gz", package="CRAVE")
    gene_file <- dirname(gene_file)   
  }

  if (op$no_gene_file) {
    op$gene_file <- NULL
  } else {
    op$gene_file  <- gene_file
  }
  op$geno_file  <- geno_file
  op$pheno_file <- pheno_file
  op$out_gene   <- out_gene
  op$out_snp    <- out_snp

  # Check the files
  if (check.files(c(geno_file, pheno_file))) stop()
  if (!op$no_gene_file) {
    if (check.files(gene_file)) stop()
  }

  # Check the output file can be opened in write mode
  temp <- file(out_gene, open="w")
  close(temp)
  temp <- file(out_snp, open="w")
  close(temp)

  
  # Check options
  temp <- op$max_maf
  if ((temp <= 0) || (temp > 0.5)) stop("ERROR: max_maf must be between 0 and 0.5")
  temp <- op$max_miss_rate
  if ((temp < 0) || (temp > 1)) stop("ERROR: max_miss_rate must be between 0 and 1")
  temp <- op$pheno_miss
  if (nchar(temp) != 1) stop("ERROR: pheno_miss must be a single character")
  temp <- op$nperm
  if (temp < 5) stop("ERROR: nperm must be at least 5")
  temp <- op$adapt_maxperm
  if (temp < 0) stop("ERROR: adapt_maxperm must be >= 0")
  temp <- op$adapt_tol
  if (temp <= 0) stop("ERROR: adapt_tol must be positive")
  temp <- op$das_alpha
  if ((temp <= 0) || (temp >= 1)) stop("ERROR: das_alpha must be between 0 and 1")
  temp <- op$threshold_alpha
  if ((temp <= 0) || (temp >= 1)) stop("ERROR: threshold_alpha must be between 0 and 1")
  temp <- op$bm_n
  if (temp < 1) stop("ERROR: bm_n must be at least 1")
  temp <- op$print
  if (temp < 0) stop("ERROR: print must be an integer >= 0")
  temp <- op$vs
  if (temp < 0) stop("ERROR: vs must be >= 0")
  temp <- op$store_perms
  if (!(temp %in% 0:1)) stop("ERROR: store_perms must be 0 or 1")
  temp <- op$gene_kb
  if ((temp < 0) || (temp > 1000)) stop("ERROR: gene_kb must be between 0 and 1000")
 

  tests <- as.character(op$tests)
  tests <- toupper(removeWhiteSpace(tests))
  test  <- unique(tests)
  all   <- c("ALL", "BM", "CMC", "DAS", "FI", "HOI", "HOT", "MDF", "MDP", "RO", "RV", "STO", "STP", "SUM", "TH", 
            "C1", "C2", "C3", "C4", "C5",
            "R1", "R2", "R3", "R4", "R5")
  all   <- toupper(all)
  temp  <- !(tests %in% all)
  if (any(temp)) {
    print(tests[temp]) 
    stop("ERROR: The above tests are not valid test names")
  }
  rm(tests, all)
  gc()

  temp <- op[["pheno_sep", exact=TRUE]]
  if (!is.null(temp)) {
    if (length(temp) > 1) stop("ERROR: pheno_sep must be 1 character")
    if (temp == "\t") op$pheno_sep <- "t"
  }
  temp <- op[["geno_sep", exact=TRUE]]
  if (!is.null(temp)) {
    if (length(temp) > 1) stop("ERROR: geno_sep must be 1 character")
    if (temp == "\t") op$geno_sep <- "t"
  }
  temp <- op[["weight_sep", exact=TRUE]]
  if (!is.null(temp)) {
    if (length(temp) > 1) stop("ERROR: weight_sep must be 1 character")
    if (temp == "\t") op$weight_sep <- "t"
  }

  # Check for a log file
  log_file <- op[["log_file", exact=TRUE]]
  log_flag <- !is.null(log_file)
  #if ((log_flag) && (op$print == 0)) op$print <- 2
  #if (op$print) {
  #  if (log_flag) {
  #    temp <- file(log_file, open="w")
  #    close(temp)
  #  } else {
  #    # Create a log file
  #    log_file <- paste(out_file, ".LOG", sep="")
  #    op$log_file <- log_file
  #  }
  #}
   
  # Check for a user-defined R function 
  user_f   <- op[["user_func", exact=TRUE]]
  userFlag <- !is.null(user_f)
  if (userFlag) {
    n_f <- length(user_f)
    if (n_f > 5) stop("ERROR: Only a maximum of 5 user-defined R functions is allowed")
    for (i in 1:n_f) {
      if (!is.function(eval(parse(text=user_f[i])))) {
        temp <- paste("ERROR: ", user_f[i], " is not a function", sep="")
        stop(temp) 
      }
    }
    main_f <- list(crave_user_func)
    op$tests <- c(op$tests, paste("r", 1:n_f, sep=""))
  } else {
    main_f <- list()
    user_f <- ""
    n_f    <- 0  
  }
  op$user_func <- NULL
  print_f <- list(crave_print)

  op$nperm <- formatC(op$nperm, format="d")
  op$adapt_maxperm <- formatC(op$adapt_maxperm, format="d")

  nop  <- length(op)
  nm   <- names(op)
  args <- character(nop+1)
  args[1] <- "crave"
  index <- 2
  for (i in 1:nop) {
    op.name <- nm[i]
    op.val  <- op[[i]]
    #len     <- length(op.val)
    if (op.name == "notGroupedByChr") {
      if (op.val) {
        args[index] <- "--notGroupedByChr"
        index <- index + 1
      }
    } else if (op.name == "no_gene_file") {
      if (op.val) {
        args[index] <- "--no_gene_file"
        index <- index + 1
      }
    } else {
      op.val <- paste(op.val, collapse=",", sep="")
      op.val <- paste(op.name, "=", op.val, sep="")
      args[index] <- op.val
      index <- index + 1 
    }
  }
  args     <- args[1:(index-1)]
  nargs    <- as.integer(length(args))
  userFlag <- as.integer(userFlag)
  retFlag  <- as.integer(1)
  n_f      <- as.integer(n_f)
  if (op$print > 7) print(args)

  ret <- .C("crave", nargs, args, userFlag, main_f, print_f, user_f, n_f, retFlag=retFlag)
  if (ret$retFlag) {
    print(ret)
    stop("ERROR: in calling C function crave")
  }

  NULL

} #END: crave


# Function to assign a default value to an element in a list
default.list <- function(inList, names, default, error=NULL,
                         checkList=NULL) {

  # inList      List
  # names       Vector of names of items in inList
  # default     List of default values to assign if a name is not found
  #             The order of default must be the same as in names.
  # error       Vector of TRUE/FALSE if it is an error not to have the
  #             name in the list. 
  #             The default is NULL
  # checkList   List of valid values for each name.
  #             Use NA to skip a list element.
  #             The default is NULL

  n1 <- length(names)
  n2 <- length(default)
  if (n1 != n2) stop("ERROR: in calling default.list")

  if (is.null(error)) {
    error <- rep(0, times=n1)
  } else if (n1 != length(error)) {
    stop("ERROR: in calling default.list")
  }

  if (!is.null(checkList)) {
    if (n1 != length(checkList)) stop("ERROR: in calling default.list")
    checkFlag <- 1
  } else {
    checkFlag <- 0
  } 

  if (is.null(inList)) inList <- list()

  listNames <- names(inList)
  for (i in 1:n1) {
    if (!(names[i] %in% listNames)) {
      if (!error[i]) {
        inList[[names[i]]] <- default[[i]]
      } else {
        temp <- paste("ERROR: the name ", names[i], " was not found", sep="")
        stop(temp)
      }
    } else if (checkFlag) {
      temp <- checkList[[i]]
      if (!all(is.na(temp))) {
        if (!all(inList[[names[i]]] %in% checkList[[i]])) {
          temp <- paste("ERROR: the name '", names[i], 
                      "' has an invalid value", sep="")
          stop(temp)
        }
      }
    }
  }

  inList

} # END: default.list

# Function to check if a file exists
check.files <- function(files) {
  # files   Character vector of files
  
  temp <- !(file.exists(files))
  if (any(temp)) {
    ret <- 1
    files <- files[temp]
    for (file in files) print(paste("The file ", file, " does not exist", sep=""))
  } else {
    ret <- 0
  }
  
  ret

} # END: check.files

removeWhiteSpace <- function(str, leading=1, trailing=1) {

  if ((leading) && (trailing)) {
    ret <- gsub("^\\s+|\\s+$", "", str, perl=TRUE)
  } else if (leading) {
    ret <- gsub("^\\s+", "", str, perl=TRUE)
  } else if (trailing) {
    ret <- gsub("\\s+$", "", str, perl=TRUE)
  } else {
    ret <- str
  }

  ret

} # END: removeWhiteSpace
