/***************************************************************
                                                             
                    CRAVE                  
                                                              
 This file is distributed under the GNU General Public        
 License, Version 2. See the file COPYING.txt for more    
 details                                                      
                                                              
****************************************************************/

/* History 
   Jun 30 2011 Use options INV_ADD and INV_DIV when a correlation matrix is singular
   Jul 29 2011 Add HOI test. Identify tests by first 3 chars.
   Aug 16 2011 In HOI test, treat inverse correlation matrix as the identity
   Nov 01 2011 Change in permutations:
                Let R be the test statistic from the observed data
                Let O be the test statistic from a permuted dataset, but using the correlation matrix from the original data
                (note O still uses the new values of variance in the deltas).
                Rule: if abs(R-O)/O < 0.3, generate the correlation matrix from the permuted data, and recalculate the test statistic
                using that matrix. Otherwise just use O.
   Dec 20 2011 Update hotelling test to use the new variances for each permutation
   May 02 2012 Free each gene after permutations have been run. Check for options used more than once.
   May 03 2012 Add code for the effect size of each snp.
               Do not update covariance matrix if all tests were: ROVER, STPOSZ, FISHER, STO, HOI
   May 07 2012 Add min_MAF option
   May 08 2012 Add flag gf_alloc to indicate when memory was allocated for gene_file
   May 30 2012 Add new code to compute the effect size
   May 31 2012 Add option to add columns to the output gene file
   Jun 07 2012 Allow for compressed genotype files
   Jun 19 2012 Use getVarNum function for determining the variable number in the pheno file 
   Sep 28 2012 Conserve memory by running permutations when current location > gene end location
   Oct 01 2012 Partition hg files
   Oct 16 2012 Add code for non-synonomous SNPs
   Nov 06 2012 Make tests more efficient
   Nov 08 2012 Add option to save matrix of permutations (for a cc study)
   Jan 09 2013 Make MDF (SRT) test the default and add an option gene_kb for +/- N kb upstream/downstream
   Jan 14 2013 Allow for test SKAT and CAL (MDF test)
   Jan 15 2013 Add code to sort output files
*/

#ifndef __CRAVE_H__
#define __CRAVE_H__

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>

/* For R package, set RFLAG to 1 */
#define RFLAG 1

#if RFLAG
#include <R.h>
#include <R_ext/Memory.h>
#include <Rmath.h>
#include <Rinternals.h>
#define WITH_ZLIB 1
#endif

#ifndef WITH_ZLIB
#define WITH_ZLIB 0
#endif

#if WITH_ZLIB
  #if RFLAG
  #include "./zlib.h"
  #else
  #include <zlib.h>
  typedef voidp gzFile;
  #endif
#else
typedef FILE * gzFile;
#endif

#ifndef DEVEL_FLAG
#define DEVEL_FLAG 0
#endif

#define VERSION "0.0.2"
#define CHECK_MEM(obj) if (obj == NULL) {printf("ERROR: allocating memory \n"); exit(0);}
#define MAX( a, b ) ( ((a) > (b)) ? (a) : (b) )
#define MIN( a, b ) ( ((a) < (b)) ? (a) : (b) )
#define MISS_GENO -1
#define MISS_PVALUE -9999.0
#define MAX_CHR_LEN 20
#define MAX_GENE_LEN 100
#define END_OF_LINE '\n'
#define CAR_RETURN '\r'
#define END_OF_STR '\0'
#define MYPI 3.141592653589793
#define ONE_OVER_SQRT_2PI 0.3989422804014327
#define MAX_ID_LEN 100
#define ALL_MISS_GENO 9
#define ERROR_GENO -9
#define ERROR_CHR 0
#define MAX_FOUND 25
#define MAX_CHR_NUMBER 27
#define ALMOST_ZERO 1e-16
#define NUM_ZERO 1e-100
#define DOUBLE_MISS -999999999.0
#define ERF_A 0.1400122886866665
#define ERF_2_PIA 4.54688497944829
#define SQRT2 1.414213562373095
#define LOGREG_MAXIT 25
#define LOGREG_EPS 1e-4
#define LOGREG_ZERO 1e-6
#define LOGREG_ONE 0.99999
#define LOGREG_FUZZ 1e-20
#define ERROR_SINGULAR_MATRIX 1
#define ERROR_MAXITER 2
#define ERROR_NO_NONMISS_GENO 1
#define ERROR_VARIANCE 3
#define ERROR_EFFECT_SIZE -9999.0
#define VCF 1
#define TPED 2
#define N01_CDFINV_01 -2.32634787404084
#define N01_CDFINV_05 -1.644853626951473
#define N01_CDFINV_10 -1.281551565544601
#define N01_CDFINV_99 2.32634787404084
#define N01_CDFINV_95 1.644853626951473
#define N01_CDFINV_90 1.281551565544601
#define MAXCORRSNP 0.95
#define PRINT_GENO_FILE 5
#define PRINT_PHENO_FILE 4
#define PRINT_WEIGHT_FILE 5
#define PRINT_INVCORR 2
#define PRINT_NEWCORR 7
#define ADD_NPERM_COLS 1
#define ADD_NGTOBS_COLS 2
#define MAX_GT_BUFFER 10000

#define OBS_FUZZ 1.0e-10
#define N_TESTS 24
#define N_NONUSER_TESTS 14
#define TEST_BM 0
#define TEST_CMC 1
#define TEST_DAS 2
#define TEST_FISHER 3
#define TEST_HOTINDEP 4
#define TEST_HOTELLING 5
#define TEST_SRT 6
#define TEST_SRTPOS 7
#define TEST_ROVER 8
#define TEST_ROVERV 9
#define TEST_STOUFFER 10
#define TEST_STPOSZ 11
#define TEST_SUM 12
#define TEST_THRESHOLD 13
#define TEST_C1 14
#define TEST_C2 15
#define TEST_C3 16
#define TEST_C4 17
#define TEST_C5 18
#define TEST_R1 19
#define TEST_R2 20
#define TEST_R3 21
#define TEST_R4 22
#define TEST_R5 23

typedef unsigned char binary;
typedef signed char genoType;
typedef int locType;
typedef double yType;
typedef unsigned char testType;

FILE *FPT_LOG_FILE;

struct filefpt {
  FILE *fpt;
  gzFile gzfpt;
  binary gzFlag;
  int buffersize;
  binary header;
  int nrow;
};
typedef struct filefpt FILEFPT;

/* Structure to save info for sorting files later */
struct out_str {
  int n_snp;
  double pvalue;
  char *gene_row;
  char **snp_row;
  struct out_str *next;
};
typedef struct out_str OUT_STR;


struct test_str0 {
  int which;    /* which test */
  double *obs;   /* observed test */
  int nperm;    /* total number of permutations */
  double pvalue;
  binary cont;  /* 0 or 1 to continue permuting */
  double *test;
  int len;
  int *n_gt_obs;  /* Number of tests > obs */
};
typedef struct test_str0 TEST_STR;

struct snp_str {
  double weight;
  genoType *G;
  binary G_equal_2;
  binary mono_control;
  double maf;
  int n0, n1, n2, nmiss;
  int *id1, *id2, *idmiss;
  double sqrtN;
  locType location;
  double effect_size;
  struct snp_str *next;  
};
typedef struct snp_str SNP_STR;

struct gene_str {
  int n_snps;
  int n_mono_control;
  char *name;  
  binary G_equal_2;
  int start;
  int end;
  SNP_STR *first;
  SNP_STR *last;
  struct gene_str *next;  
  struct gene_str *previous; 
};
typedef struct gene_str GENE_STR;

struct chr_str {
  int n_genes;
  GENE_STR *first;
  GENE_STR *last;
};
typedef struct chr_str CHR_STR;

/* Structure to hold the gene table for a chr */
struct gene_table {
  int len;
  locType *loc_start;
  locType *loc_stop;
  char **gene_name;
  locType *codon_start;
  locType *codon_stop;
  int *n_exon;
  locType **exon_start;
  locType **exon_stop;
  int nonSyn_n;
  locType *nonSyn_loc;
};
typedef struct gene_table GENE_TABLE;

struct info_str {
 
  /* Structure to hold data */
  CHR_STR **chr;
  int n_chr_str;
  binary chr0Flag;
  int chr_str_index;
  int errorFlag;

  /* Output file */
  char *out_file;
  FILE *out_fpt;
  int n_chr_output;
  char *log_file;
  char *out_snp;
  FILE *out_fpt_snp;
  OUT_STR *out_save;  /* Info about each gene to save for sorting the output data sets */
  binary sort_out;    /* Flag to sort output files by (first) gene p-value */

  /* Phenotype file */
  char *pheno_file;
  int ncovs;
  binary *use_covars;
  int n_use_covars;
  char *covar_str;
  double **covmat;
  double **temp_covmat;
  int nsubs;
  int pheno_nrow;
  int pheno_ncol;
  binary *use_pheno;
  yType *CC;
  yType *temp_CC;
  binary CC_flag;
  char **pheno_ids;
  int id_len;
  int n_case;
  int n_control;
  char pheno_miss;
  int response_col;
  int id_col;
  int case_value;
  int control_value;
  char *response_var;
  char *id_var;

  /* Pheno options */
  char pheno_sep;
  long pheno_buffer;
  binary pheno_header;
  binary inc_sub_flag;
  binary exc_sub_flag;
  char *sub_file;   /* File containing list of samples to include/exclude */

  /* Genotype file */
  FILEFPT gfpt;
  char *geno_file;
  char **geno_cols;
  binary *use_geno;
  char *current_chr;
  char *last_chr;
  int current_chr_num; /* Numeric value of chr */
  int last_chr_num; 
  int n_chr;
  char *current_gene;
  char *last_gene;
  int current_gene_len;
  int last_gene_len;
  int geno_ncols;
  int geno_line_num;
  int gene_var_len;
  genoType *genotypes;
  int n_genotypes;
  locType current_pos;
  locType last_pos;
  double current_weight;
  binary groupedByChr;
  int file_type; /* 1 = vcf, 2 = tped */
  binary AF_flag;   /* For alternate allele freq in vcf file */
  binary ACAN_flag; /* For alternate allele counts in vcf file */

  /* Geno file options */
  char geno_sep; 
  long geno_buffer;
  char *gene_var;
  char *changesAA_var;
  int changesAA_var_len;
  char *functionalClass_var;
  int functionalClass_var_len;
  int gene_size;
  int chr_size; 
  double max_maf, min_maf;
  char allele_miss;  /* Missing value for alleles in TPED file */
  double max_missRate;
  char *chrStr;
  binary includeChr[MAX_CHR_NUMBER];
  int includeChrMax;
  binary inc_chr_flag;
  binary exc_chr_flag;
  char **chr_vec;
  int chr_vec_n;
  binary inc_snp_flag;
  binary exc_snp_flag;
  char *snp_file;   /* File containing list of snps to include/exclude */
  char **snp_vec;
  int snp_vec_n;
  binary sortedByLoc;  /* 0 or 1 if genotype file is sorted by location within each chr */

  /* Gene Table */
  GENE_TABLE **gene_table;
  binary codonFlag;
  binary exonFlag;
  binary nonSynFlag; /* Flag for including non-synonomous SNPs */
  char *hg_dir;   /* Directory for partitioned hg files */
  char *gene_file;
  binary useGeneTable;
  long gene_buffer;
  int *genes_found;  /* Genes in the table for a given location */
  int n_found;
  int max_chr_number;
  int *used_chr; /* To keep track if the tped file is grouped by chr */
  char **chr_alias; /* Chromosome alias (for VCF files) */
  char *chr_alias_file;  /* File for chr aliases */
  binary inc_gene_flag;
  binary exc_gene_flag;
  char *genelist_file;   /* File containing list of snps to include/exclude */
  char **gene_vec;       /* Genes to include/exclude */
  int gene_vec_n;
  int current_gene_start;
  int current_gene_end;
  int gene_kb; /* Number of kilobases upstream/downstream for defining gene regions */

  /* Weight file */
  char *weight_file;
  char weight_sep;
  binary weight_header;
  char **weight_snps;
  double *weight_wgts;
  int weight_n;
  FILE *weight_fpt;
  int weight_type;

  /* For test stats */
  double **geno_cov;
  double **geno_invcor;
  double **geno_cor;
  binary geno_cov_flag;
  binary geno_invcor_flag;
  binary geno_cor_flag;
  double *geno_sdG; /* sqrt(diag(geno_cov)) */
  double *geno_varG; /* diag(geno_cov) */
  double *Y_norm;
  double *delta;
  char *test_str;
  binary test_flag[N_TESTS];
  int n_test;
  double das_alpha;
  double qnorm_alpha;  /* quantile of alpha */
  int bm_n;
  double test_cc_const;
  double **gstar_invcor;
  int gstar_ncat;
  int cmc_ncat;
  double threshold_alpha;
  double threshold_phi;
  TEST_STR **tests;
  binary user_def_flag;
  double rover_t;
  double *user_delta;
  double *user_cov;
  double *user_cor;
  double *user_invcor;
  double *delta2;  /* For singular correlation matrices */
  int delta2_n;  /* Length of delta2 */
  double **geno_invcor2; /* Computed when geno_invcor is singular */
  GENE_STR *GENE2;  /* GENE structure containing subset of SNPs in a gene */
  double *geno_sdG2; /* sqrt(diag(geno_cov2)) */
  testType test_order[N_TESTS];
  double update_cor;
  binary SKAT_flag; /* Flag is user specified SKAT test */
  binary CAL_flag; /* Flag is user specified CAL test */


  /* For permutations */
  int random_seed;
  int nperm;
  binary adapt_flag;
  double adapt_tol;
  int adapt_maxiter;
  int adapt_maxperm;
  binary storePermFlag;  /* Flag to store the matrix of permutations for a cc study */
  double **storePermMat; /* Matrix of permutations */

  char *out_perm;
  FILE *perm_fpt;

  binary print;
  long maxrows;
  binary snp_in_mult_genes;
  int n2_maf;
  binary useControlsForV;
  binary *controls;
  char *id_str;
  binary store_snp_flag;
  double INV_ADD;
  double INV_DIV;
  double vs;  /* variance stabilization */
  binary effect_size;
  binary gf_alloc;
  char add_gcols;
  binary finalChr;   /* Flag to set to 1 when chrPermsOutput is called */
  binary newChrFlag; /* Flag to set to 1 when a new chr is reached */
  binary weightFlag; /* Flag that indicates if the weight= option was specified */

};
typedef struct info_str INFO;



void bvecinit(binary *, int, binary);
int matchCharVecs(char **, binary *, int, char **, int, binary *, binary *, int *); 
int findStrInVec(char *, char **, int);
char ** strIntoVec0(char *, int, char *);
char ** strIntoVec(char *, char, int *);
int strIndexInVec(char *, char **, int);
int strIndexInVec2(char *, char **, int, binary *);
int count_strinstr(char *, char *);
char * extractStr2(char *, char *, char *, char *);
int getStr2delim(char *, int, char *, char *, char, char *);
FILE * open_file(char *, char *);
int getNcols(char *, char *, long);
int getNrows(char *, long);
void checkEndOfLine(char *, int, long, char *);
void print_info(INFO *, int);
void print_help(int);
void print_chrStr(INFO *);
void print_option(int);
int binSearch_int(int *, int , int);


char * next_token(char *, long, char);
int copy_token(char *, char *, char, long);
char ** cVec_alloc(int, int);
char * c0Vec_alloc(int);
double * dVec_alloc(int, int, double);
int * iVec_alloc(int, int, int);
int **intMatrix_alloc(int, int);
double **dblMatrix_alloc(int, int);
binary **binMatrix_alloc(int, int);
void matrix_free(void **, int);
void free_info(INFO *);
void free_chr(INFO *, int);

void init(INFO *, char **, int);
void getPheno2(INFO *);
void getCols(INFO *);
void getCols_tped(INFO *);
void getCols_vcf(INFO *);
void AFACAN_vcf(INFO *);
int count_char(char *, char);
void checkCols_vcf(INFO *);
void update_info(INFO *);
void read_geno(INFO *);
int parseLine_vcf(char *, INFO *);
int parseLine_tped(char *, INFO *);
void open_out(INFO *);
long buffer_size(char *, int, int, FILEFPT *);
int check_pos(char *, INFO *);
void init_snp(SNP_STR *, genoType *, int, double, double, int, locType);
void copyGeneStr(GENE_STR *, binary *, int, GENE_STR *, int);
void free_gene(GENE_STR *, binary);
void print_geneLink(GENE_STR *, int, int);

void runPermsOutput(INFO *);
void chrPermsOutput(INFO *);
double c1(double *, double **, double **, double **, int);
double c2(double *, double **, double **, double **, int);
double c3(double *, double **, double **, double **, int);
double c4(double *, double **, double **, double **, int);
double c5(double *, double **, double **, double **, int);

double pvalue_normal(double, double, double);
double pvalue_normal01(double);
int genoCounts(genoType *, int, int *);
void switch_genos(genoType *, int);
void add_geno(genoType *, INFO *, double);
int getChrNum(char *, int, int, char **);
int getGeneFromTable(INFO *);
void readGeneTable(INFO *);
void X_Xp_miss(double **, int, int, double, double **);
int EY_X(double **, int, int, double *, binary, double *);
void norm_dvec(double *, int, int, double *); 
void DV_V_DV(double **, int, double **);
void getY_norm(INFO *);
int cov_inv(double **, int, double **);
int tests_init(INFO *, GENE_STR *);
void compute_tests2(INFO *, GENE_STR *, binary);
void permute_tests(INFO *, GENE_STR *);
void mean_var_dmat(double **, int, int, double *, double *);
double min_dvec(double *, int);
int min_ivec(int *, int);
double normal01_cdfinv(double);
void getDirName(char *, int, char *);
void copySqMat_exclude(double **, binary *, int, double **, int);
void printf_s(char *, char *);
void printf_i(char *, int);
void printf_d(char *, double);
void printf_c(char *, char);
int getIntsFromStr(char *, int, int, binary *);
int getHeader(char *, char, int, long);
void quicksort_str(int, int, char **, int *);
char ** readIds(char *, long, int, int, int *);
int removeStr(char *, char **, int, binary);
char getDelimiter(char *, long, FILEFPT *);
char ** getColNames(char *, char, long, int, int *);
int checkVarsInSortVec(char *, char, char **, int, binary *, int *);
int getChrsFromStr(char *, int, binary *, char **, int);
double getWeight(INFO *, int);
int beta_normal(double **, int, int, double *, double *, double **);
int beta_logReg(double **, int, int, double *, double *, double **);
void effect_size_all(INFO *, GENE_STR *);
void effect_size(INFO *, GENE_STR *);
int getVarNum(char *, char **, int, int);
int gt_exclude_snp(INFO *);
void getPermMat(INFO *);


void my_fopen(char *, FILEFPT *); 
void my_fclose(FILEFPT *); 
char * my_fgets(char *, size_t, FILEFPT *);
int is_gz_file(char *);

#if RFLAG
double user_r_main(INFO *, GENE_STR *, int);
int print_r_main(char *, char *, double *, int *, char *);
#endif

/* REMOVE */
#if DEVEL_FLAG
void print_gVec(genoType *, int, char *);
void print_iVec(int *, int, char *);
void print_cVec(char **, int, char *);
void print_bVec(binary *, int, char *);
void print_dMat(double **, int, int, char *);
void print_dVec(double *, int, char *);
void print_tests(INFO *, int);
#endif

#endif







