/***************************************************************
                                                             
                    CRAVE                  
                                                              
 This file is distributed under the GNU General Public        
 License, Version 2. See the file COPYING.txt for more    
 details                                                      
                                                              
****************************************************************/

#include "./crave.h"

/* Function to create and initialize an out structure */
static OUT_STR * outstr_init(n_snp, pval)
int n_snp;
double pval;
{
  OUT_STR *ret;

  ret = (OUT_STR *) malloc(sizeof(OUT_STR));
  CHECK_MEM(ret);

  ret->n_snp    = n_snp;
  ret->pvalue   = pval;
  ret->gene_row = NULL;
  ret->snp_row  = NULL;
  ret->next     = NULL;

  return(ret);

} /* END: outstr_init */

/* Function to free an out structure */
static void outstr_free(str)
OUT_STR *str;
{
  if (str->gene_row) free(str->gene_row);
  if (str->snp_row) matrix_free((void **)str->snp_row, str->n_snp);
  free(str);

} /* END: outstr_free */

/* Function to free all out structures */
static void outstr_free_all(str)
OUT_STR *str;
{
  OUT_STR *out, *next;

  out = str;
  while (1) {
    if (!out) break;
    next = out->next;
    outstr_free(out);
    out = next;
  }

} /* END: outstr_free_all */

/* Function to allocate a gene_table structure */
static GENE_TABLE * gt_alloc(n, info)
int n;
INFO *info;
{
  GENE_TABLE *ret;
  int i;
  char **pc;

  ret = (GENE_TABLE *) malloc(sizeof(GENE_TABLE));
  CHECK_MEM(ret);
  ret->len         = 0;
  ret->loc_start   = NULL;
  ret->loc_stop    = NULL;
  ret->gene_name   = NULL;
  ret->codon_start = NULL;
  ret->codon_stop  = NULL;
  ret->exon_start  = NULL;
  ret->exon_stop   = NULL;
  ret->n_exon      = 0;
  ret->nonSyn_n    = 0;
  ret->nonSyn_loc  = NULL;

  if (!n) return(ret);
  ret->len = n;

  ret->loc_start = (locType *) malloc(n*sizeof(locType));
  CHECK_MEM(ret->loc_start);
  ret->loc_stop = (locType *) malloc(n*sizeof(locType));
  CHECK_MEM(ret->loc_stop);
  ret->gene_name = (char **) malloc(n*sizeof(char *));
  CHECK_MEM(ret->gene_name);
  for (i=0, pc=ret->gene_name; i<n; i++, pc++) *pc = NULL;

  if (info->codonFlag) {
    ret->codon_start = (locType *) malloc(n*sizeof(locType));
    CHECK_MEM(ret->codon_start);
    ret->codon_stop = (locType *) malloc(n*sizeof(locType));
    CHECK_MEM(ret->codon_stop);
  }
  if (info->exonFlag) {
    ret->n_exon = (int *) malloc(n*sizeof(int));
    CHECK_MEM(ret->n_exon);
    ret->exon_start = (locType **) malloc(n*sizeof(locType *));
    CHECK_MEM(ret->exon_start);
    ret->exon_stop = (locType **) malloc(n*sizeof(locType *));
    CHECK_MEM(ret->exon_stop);
  }

  return(ret);

} /* END: gt_alloc */

/* Function to free a gene_table structure */
static void gt_free(obj)
GENE_TABLE *obj;
{
  if (!obj) return;
  if (obj->loc_start) free(obj->loc_start);
  if (obj->loc_stop) free(obj->loc_stop);
  if (obj->gene_name) matrix_free((void **) obj->gene_name, obj->len);
  if (obj->codon_start) free(obj->codon_start);
  if (obj->codon_stop) free(obj->codon_stop);
  if (obj->exon_start) matrix_free((void **) obj->exon_start, obj->len);
  if (obj->exon_stop) matrix_free((void **) obj->exon_stop, obj->len);
  if (obj->n_exon) free(obj->n_exon);
  if (obj->nonSyn_loc) free(obj->nonSyn_loc);
  
  free(obj);
  
} /* END: gt_free */

/* Function to free all gene tables */
static void gt_free_all(info)
INFO *info;
{
  int i;
  GENE_TABLE **GT=info->gene_table;
  
  for (i=0; i<MAX_CHR_NUMBER; i++) {
    gt_free(GT[i]);
    GT[i] = NULL;
  }

} /* END: gt_free_all */

/* Function to free the test structure */
static void free_tests(info)
INFO *info;
{
  TEST_STR **tests=info->tests, *tst;
  int i;  

  for (i=0; i<info->n_test; i++) {
    tst = tests[i];
    if (tst->test != NULL) free(tst->test);
    if (tst->obs != NULL) free(tst->obs);
    if (tst->n_gt_obs != NULL) free(tst->n_gt_obs);
    free(tst);
  }
  free(tests);
  info->tests = NULL;
  
} /* END: free_tests */

/* Function to initialize the test structure */
static void init_tests(info)
INFO *info;
{
  int i, len=info->n_test, nn, which;
  TEST_STR **tests, *tst;
  testType *tt=info->test_order;

  info->tests = (TEST_STR **) malloc(len*sizeof(TEST_STR *));
  CHECK_MEM(info->tests);
  tests = info->tests;

  for (i=0; i<len; i++) {
   
      tests[i] = (TEST_STR *) malloc(sizeof(TEST_STR));
      CHECK_MEM(tests[i]);
      which = tt[i];
      tst = tests[i];
      tst->nperm    = 0;
      tst->cont     = 0;
      tst->n_gt_obs = 0;
      tst->which    = which;
      if (which != TEST_BM) {
        tst->len = 0;
        nn = 1;
      } else {
        nn = info->bm_n;
        tst->len = nn;
      }

      tst->obs = (double *) malloc(nn*sizeof(double));
      CHECK_MEM(tst->obs);
      tst->test = (double *) malloc(nn*sizeof(double));
      CHECK_MEM(tst->test);
      tst->n_gt_obs = (int *) malloc(nn*sizeof(int));
      CHECK_MEM(tst->n_gt_obs);
  }

} /* END: init_tests */

/* Function to initialize the chr array */
static void alloc_chr_str(info)
INFO *info;
{
  int n, i;

  if (!info->groupedByChr) {
    n = info->max_chr_number;
    info->chr0Flag = 0;
  } else {
    n = 1;
    info->chr0Flag = 1;
    info->chr_str_index = 0;
    info->n_found = 1;
  }
  info->n_chr_str = n;
  info->chr = (CHR_STR **) malloc(n*sizeof(CHR_STR *));
  CHECK_MEM(info->chr);
 
  for (i=0; i<n; i++) {
    info->chr[i] = (CHR_STR *) malloc(sizeof(CHR_STR));
    CHECK_MEM(info->chr[i]);
    info->chr[i]->n_genes = 0;
    info->chr[i]->first = NULL;
    info->chr[i]->last = NULL;
  }

} /* END: alloc_chr_str */

/* Function to initialize a snp link */
void init_snp(str, geno, ngeno, wgt, maf, flag, pos)
SNP_STR *str;
genoType *geno;
int ngeno;
double wgt, maf;
int flag; /* 0 or 1 to get ids for 1 and 2 */
locType pos;  /* position */
{
  int i, freq[3], n1, n2, temp;    
  genoType *p1, *p2, gg;

  str->weight   = wgt;
  str->maf      = maf;
  str->location = pos;
  str->G = (genoType *) malloc(ngeno*sizeof(genoType));
  CHECK_MEM(str->G);
  str->effect_size = ERROR_EFFECT_SIZE;

  /* Get the frequency counts */
  if (flag) {
    freq[0] = freq[1] = freq[2] = 0;
    for (i=0, p1=str->G, p2=geno; i<ngeno; i++, p1++, p2++) {
      gg = *p2;
      *p1 = gg;
      if (gg != MISS_GENO) freq[gg]++;
    }
    str->n0     = freq[0];
    n1          = freq[1];
    n2          = freq[2];
    str->n1     = n1;
    str->n2     = n2;
    str->id1    = NULL;
    str->id2    = NULL;
    str->idmiss = NULL;

    temp = freq[0] + n1 + n2;
    str->sqrtN = sqrt(temp);
    str->nmiss = ngeno - temp;
  } else {
    str->id1    = NULL;
    str->id2    = NULL;
    str->idmiss = NULL;
    str->n0 = str->n1 = str->n2 = str->nmiss = -1;
    str->sqrtN = -1.0;
    for (i=0, p1=str->G, p2=geno; i<ngeno; i++, p1++, p2++) *p1 = *p2;
  }
  str->next = NULL;

} /* init_snp */

/* Function to free a snp link */
static void free_snp(str)
SNP_STR *str;
{
  free(str->G);
  if (str->id1 != NULL) free(str->id1);
  if (str->id2 != NULL) free(str->id2);
  free(str);

} /* END: free_snp */

/* Function to initialize a gene link */
static void init_gene(str, gene, snpStr, prevGene, flag, start, end)
GENE_STR *str, *prevGene;
char *gene;
SNP_STR *snpStr;
binary flag;
locType start, end;
{
  int len;

  str->n_snps = 1;
  str->first = snpStr;
  str->last = snpStr;
  str->next = NULL;
  str->previous = prevGene;
  str->start = start;
  str->end   = end;
  if (flag) {
    str->name = gene;
  } else {
    len = strlen(gene) + 1;
    str->name = (char *) malloc(len*sizeof(char));
    CHECK_MEM(str->name);
    strcpy(str->name, gene);
  }
  
} /* init_gene */

/* Function to free a gene link */
void free_gene(str, flag)
GENE_STR *str;
binary flag;
{
  int i, n;
  SNP_STR *snp, *next;

  snp = str->first;
  next = snp->next;
  n = str->n_snps;
  for (i=1; i<=n; i++) {
    free_snp(snp);
    if (i < n) {
      snp = next;
      next = snp->next;
    }
  }
  if (!flag) free(str->name);
  free(str);

} /* END: free_gene */

/* Function to free the chr structure */
void free_chr(info, flag)
INFO *info;
int flag;
{
  int i, n, j;
  CHR_STR *chr, **chr00=info->chr;
  GENE_STR *gene, *next;
  binary useGeneTab=info->useGeneTable;

  for (j=0; j<info->n_chr_str; j++) {
    chr = chr00[j];
    n = chr->n_genes;
    if (!n) continue;
    gene = chr->first;
    if (!gene) continue;
    next = gene->next;
    for (i=1; i<=n; i++) {
      free_gene(gene, useGeneTab);
      if (i < n) {
        gene = next;
        next = gene->next;
      }
    }
    chr->n_genes = 0;
    chr->first = NULL;
    chr->last = NULL;
    if (flag) free(chr);
  }
  if (flag) free(info->chr);

} /* free_chr */

/* Function to get a default buffer size */
long buffer_size(datafile, initSize, nLines, gfpt)
char *datafile;
int initSize, nLines;
FILEFPT *gfpt;
{
  int i, flag, count, flag2=0, eofFlag=0;
  long len0, len, maxlen=-1;
  char line[initSize], c1;
  size_t size;

  i = 1;
  count = 0;
  size = sizeof(line);
  my_fopen(datafile, gfpt);
  while (1) {
    len = 0;
    flag = 0;
    while(1) {
      if (my_fgets(line, size, gfpt) != NULL) {
        /* In genotype file, comment rows start with a # */
        if (line[0] != '#') flag = 1;
        len0 = strlen(line);
        len += len0;
        c1 = line[len0-1];
        if ((c1 == END_OF_LINE) || (c1 == CAR_RETURN)) {
          flag2 = 1;
          break;
        } else if (c1 == EOF) {
          eofFlag = 1;
          break;
        }
      } else {
        eofFlag = 1;
        break;
      }
    }
    if (len > maxlen) maxlen = len;
    count += flag;
    if (count > nLines) break;
    if (eofFlag) break;
    i++;
    if (i > 99999) {
      printf_s("\n***WARNING***: A default buffer size could not be determined for file:\n %s \n", datafile);
      maxlen = -1;
      break;
    }
  }
  my_fclose(gfpt);
  return(maxlen);

} /* END: buffer_size */

/* Function to add a snp */
void add_geno2(geno, info, maf)
genoType *geno;
INFO *info;
double maf;
{
  int n_genes, i, flag;
  CHR_STR *chr=info->chr[info->chr_str_index];
  GENE_STR *gene;
  SNP_STR *snp;
  char *cgene;

  /* Create a snp link */
  snp = (SNP_STR *) malloc(sizeof(SNP_STR));
  CHECK_MEM(snp);
  init_snp(snp, geno, info->nsubs, info->current_weight, maf, 1, info->current_pos);

  n_genes = chr->n_genes;
  if (n_genes) {
    /* Check if the gene is in the linked list. Start from the last link. */
    gene = chr->last;
    cgene = info->current_gene;
    flag = 0;
    for (i=0; i<n_genes; i++) {
      if (!(strcmp(gene->name, cgene))) {
        /* Add the snp to this gene */
        gene->last->next = snp;
        gene->last = snp;
        gene->n_snps++;
        flag = 1; 
        break;
      }
      gene = gene->previous;
    } /* END: for */
    if (!flag) {
      /* Add a new gene link */
      gene = (GENE_STR *) malloc(sizeof(GENE_STR));
      CHECK_MEM(gene);  
      init_gene(gene, info->current_gene, snp, chr->last, info->useGeneTable, info->current_gene_start, info->current_gene_end);
      chr->last->next = gene;
      chr->last = gene;
      chr->n_genes++;
    }
  } else {
    /* Start the linked list */
    gene = (GENE_STR *) malloc(sizeof(GENE_STR));
    CHECK_MEM(gene);  
    init_gene(gene, info->current_gene, snp, NULL, info->useGeneTable, info->current_gene_start, info->current_gene_end);
    chr->first = gene;
    chr->last = gene;
    chr->n_genes = 1;
  }
  
} /* END: add_geno2 */

/* Function to add a snp to the chr structure */
void add_geno(geno, info, maf)
genoType *geno;
INFO *info;
double maf;
{
  int n, i, *found, row, chr_num=info->current_chr_num;
  GENE_TABLE *GT;

  if (!info->chr0Flag) info->chr_str_index = chr_num;
  n = info->n_found;
  if (n == 1) {
    add_geno2(geno, info, maf);
  } else {
    GT = info->gene_table[chr_num];
    for (i=0, found=info->genes_found; i<n; i++, found++) {
      /* Copy gene from table */
      row = *found;
      info->current_gene       = GT->gene_name[row];
      info->current_gene_start = GT->loc_start[row];
      info->current_gene_end   = GT->loc_stop[row];
      add_geno2(geno, info, maf);
    }
  }

} /* END: add_geno */

/* Function to parse the test string */
static void parseTestStr(info)
INFO *info;
{
  int len=0, flag, i;
  char *str=info->test_str, test[21], *token, c1, c2, c3;
  binary *t=info->test_flag, *pb, user=0, MDF_flag=0;
  testType *tt=info->test_order, *pt;

  for (i=0, pb=t; i<N_TESTS; i++, pb++) *pb = 0;
     
  if (str != NULL) len = (int) strlen(str);

  if ((str == NULL) || (len == 0)) {
    /* Default */
    t[TEST_SRT] = 1;
    tt[0] = TEST_SRT;
  } else {
        /* Loop through the string */
        token = str;
        i = 0;
        while(1) {
          /* The copy_token was really for reading from a file, so flag=1 is ok */
          flag = copy_token(test, token, ',', 20);
          if (flag > 1) {
            printf_s("\nERROR: with tests=%s\n", str);
            exit(1);
          }

          len = (int) strlen(test);
          if (len > 1) {
            /* Check first 2 or 3 chars */
            c1 = tolower(test[0]);
            c2 = tolower(test[1]);
            if (len > 2) {
              c3 = tolower(test[2]);
            } else {
              c3 = 'X';
            }
     
            flag = 0;
            if ((c1 == 'a') && (c2 == 'l') && (c3 == 'l')) {
              /* All test stats */
              for (i=0, pb=t; i<N_NONUSER_TESTS; i++, pb++) *pb = 1;
              for (i=0, pt=tt; i<N_NONUSER_TESTS; i++, pt++) *pt = i;
              info->geno_cov_flag = 1;
              info->geno_invcor_flag = 1;
              /*break;*/
            } else if ((c1 == 's') && (c2 == 'u')) {
              if (!t[TEST_SUM]) {
                t[TEST_SUM] = 1;
                tt[i++] = TEST_SUM;
                info->geno_cov_flag = 1;
              }
            } else if ((c1 == 'h') && (c2 == 'o') && (c3 == 't')) {
              if (!t[TEST_HOTELLING]) {
                t[TEST_HOTELLING] = 1;
                tt[i++] = TEST_HOTELLING;
                info->geno_cov_flag = 1;
                info->geno_invcor_flag = 1;
              }
            } else if ((c1 == 's') && (c2 == 't') && (c3 == 'o')) {
              if (!t[TEST_STOUFFER]) {
                t[TEST_STOUFFER] = 1;
                tt[i++] = TEST_STOUFFER;
              }
            } else if ((c1 == 'f') && (c2 == 'i')) {
              if (!t[TEST_FISHER]) {
                t[TEST_FISHER] = 1;
                tt[i++] = TEST_FISHER;
              }
            } else if ((c1 == 'c') && (c2 == 'm')) {
              if (!t[TEST_CMC]) {
                t[TEST_CMC] = 1;
                tt[i++] = TEST_CMC;
                info->geno_cov_flag = 1;
                info->geno_invcor_flag = 1;
              }
            } else if ((c1 == 'm') && (c2 == 'd') && (c3 == 'f')) {
              if (!t[TEST_SRT]) {
                t[TEST_SRT] = 1;
                tt[i++] = TEST_SRT;
                MDF_flag = 1;
              }
            } else if ((c1 == 's') && (c2 == 'k')) {
              if (!t[TEST_SRT]) {
                t[TEST_SRT] = 1;
                tt[i++] = TEST_SRT;
                info->SKAT_flag = 1;
              }
            } else if ((c1 == 'c') && (c2 == 'a')) {
              if (!t[TEST_SRT]) {
                t[TEST_SRT] = 1;
                tt[i++] = TEST_SRT;
                info->CAL_flag = 1;
              }
            } else if ((c1 == 'm') && (c2 == 'd') && (c3 == 'p')) {
              if (!t[TEST_SRTPOS]) {
                t[TEST_SRTPOS] = 1;
                tt[i++] = TEST_SRTPOS;
              }
            } else if ((c1 == 'd') && (c2 == 'a')) {
              if (!t[TEST_DAS]) {
                t[TEST_DAS] = 1;
                tt[i++] = TEST_DAS;
              }
            } else if ((c1 == 'b') && (c2 == 'm')) {
              if (!t[TEST_BM]) {
                t[TEST_BM] = 1;
                tt[i++] = TEST_BM;
                info->geno_cov_flag = 1;
                info->geno_invcor_flag = 1;
              }
            } else if ((c1 == 's') && (c2 == 't') && (c3 == 'p')) {
              if (!t[TEST_STPOSZ]) {
                t[TEST_STPOSZ] = 1;
                tt[i++] = TEST_STPOSZ;
              }
            } else if ((c1 == 't') && (c2 == 'h')) {
              if (!t[TEST_THRESHOLD]) {
                t[TEST_THRESHOLD] = 1;
                tt[i++] = TEST_THRESHOLD;
                info->geno_cov_flag = 1;
                info->geno_invcor_flag = 1;
              }
            } else if ((c1 == 'r') && (c2 == 'o')) {
              if (!t[TEST_ROVER]) {
                t[TEST_ROVER] = 1;
                tt[i++] = TEST_ROVER;
              }
            } else if ((c1 == 'r') && (c2 == 'v')) {
              if (!t[TEST_ROVERV]) {
                t[TEST_ROVERV] = 1;
                tt[i++] = TEST_ROVERV;
              }
             } else if ((c1 == 'h') && (c2 == 'o') && (c3 == 'i')) {
              if (!t[TEST_HOTINDEP]) {
                t[TEST_HOTINDEP] = 1;
                tt[i++] = TEST_HOTINDEP;
              }
            }  else if ((c1 == 'c') && (c2 == '1')) {
              if (!t[TEST_C1]) {
                t[TEST_C1] = 1;
                tt[i++] = TEST_C1;
                user = 1;
              }
            } else if ((c1 == 'c') && (c2 == '2')) {
              if (!t[TEST_C2]) {
                t[TEST_C2] = 1;
                tt[i++] = TEST_C2;
                user = 1;
              }
            } else if ((c1 == 'c') && (c2 == '3')) {
              if (!t[TEST_C3]) {
                t[TEST_C3] = 1;
                tt[i++] = TEST_C3;
                user = 1;
              }
            } else if ((c1 == 'c') && (c2 == '4')) {
              if (!t[TEST_C4]) {
                t[TEST_C4] = 1;
                tt[i++] = TEST_C4;
                user = 1;
              }
            } else if ((c1 == 'c') && (c2 == '5')) {
              if (!t[TEST_C5]) {
                t[TEST_C5] = 1;
                tt[i++] = TEST_C5;
                user = 1;
              }
            } else if ((c1 == 'r') && (c2 == '1')) {
              if (!t[TEST_R1]) {
                t[TEST_R1] = 1;
                tt[i++] = TEST_R1;
                user = 1;
              }
            } else if ((c1 == 'r') && (c2 == '2')) {
              if (!t[TEST_R2]) {
                t[TEST_R2] = 1;
                tt[i++] = TEST_R2;
                user = 1;
              }
            } else if ((c1 == 'r') && (c2 == '3')) {
              if (!t[TEST_R3]) {
                t[TEST_R3] = 1;
                tt[i++] = TEST_R3;
                user = 1;
              }
            } else if ((c1 == 'r') && (c2 == '4')) {
              if (!t[TEST_R4]) { 
                t[TEST_R4] = 1;
                tt[i++] = TEST_R4;
                user = 1;
              }
            } else if ((c1 == 'r') && (c2 == '5')) {
              if (!t[TEST_R5]) {
                t[TEST_R5] = 1;
                tt[i++] = TEST_R5;
                user = 1;
              }
            } else {
                printf_s("\nERROR: test = %s is not valid\n", test);
                exit(1);
            }
          } else if (len > 0) {
            printf_s("\nERROR: test = %s is not valid\n", test);
            exit(1);
          }
       
          /* Get the next token */
          token = next_token(token, 20, ',');
          if (token == NULL) break;

        } /* END: while */
  }

  if (user) {
    info->user_def_flag = 1;
    info->geno_cov_flag = 1;
    info->geno_cor_flag = 1;
    info->geno_invcor_flag = 1;
  }

  len = 0;
  for (i=0, pb=t; i<N_TESTS; i++, pb++) {
    if (*pb) len++;
  }
  info->n_test = len;
  if (len == 0) {
    t[TEST_SRT]  = 1;
    info->n_test = 1;
    len = 1;
  }

  /* For bounded minimum test */
  if ((t[TEST_BM] == 1) && (info->bm_n == 0)) info->bm_n = 10;

  /* Allocate test structures */
  init_tests(info);

} /* END: parseTestStr */

/* Function to return the delimiter */
char get_sep(str, errStr)
char *str, *errStr;
{
  int len;
 
  len = strlen(str);

  if (!len) {
    printf_s("\n ERROR: The option %s must be a single character \n", errStr);
    exit(1);
  } else if (len == 1) {
    if (*str == 't') {
      return('\t');
    } else {
      return(*str);
    }
  } else {
    printf("\n ERROR: %s is not a valid value for %s \n", str, errStr);
    if (FPT_LOG_FILE) fprintf(FPT_LOG_FILE, "\n ERROR: %s is not a valid value for %s \n", str, errStr);
    exit(1);
  }
  
  return('\t');

} /* get_sep */

/* Function to parse the input options */
void parse_options(info, argv, argc)
INFO *info;
char **argv;
int argc;
{
  int nop=71, i, j, count=0, opNum;
  char *ops[71] = {"pheno_file=", "geno_file=", "out_gene=", "max_maf=", "nperm=", "seed=",\
                   "--help", "pheno_sep=", "print=", "reflink.geneName=", "chr_buffer=", "id_buffer=",\
                   "pheno_buffer=", "geno_buffer=",  "geno_sep=", "gene_buffer=", "maxrows=",\
                   "refseq.changesAA=", "refseq.functionalClass=", "gene_file=",\
                   "--version", "--notGroupedByChr", "allele_miss=", "maxChrNum=",\
                   "pheno_miss=", "covars=", "tests=", "das_alpha=", "bm_n=", "adapt_tol=", "adapt_maxiter=",\
                   "max_miss_rate=", "threshold_alpha=", "--useAllSubsForV", "--no_gene_file", "rover_t=",\
                   "log_file=", "include_chr=", "pheno_header=", "response=", "id=",\
                   "case_value=", "control_value=", "include_snp=", "exclude_snp=", "include_gene=", "exclude_gene=",\
                   "include_sub=", "exclude_sub=", "exclude_chr=",\
                   "weight_file=", "weight_sep=", "weight_header=", "weight_type=", "INV_ADD=", "INV_DIV=",\
                   "adapt_maxperm=", "out_snp=", "vs=", "out_perm=", "update_cor=", "effect_size=", "min_maf=",\
                   "addcols=", "geno_gz=", "sortedByLoc=", "weight=", "chr_alias_file=", "store_perms=", "gene_kb=",\
                   "sort_out="};
  binary pflag[nop], *ptrb;
  int len[nop], *ptri, temp, which, req[3]={0, 1, 2}, flag=0, helpFlag=0, ncall[nop];
  char **ptrc, *pc, *end, *opName, c1;
  long buf;
  double dval;

  /* If no arguements, print help and exit */
  if (argc <= 1) {
    print_help(0);
    for (i=0; i<2; i++) printf_s("\nERROR: the option %s must be specified \n", ops[req[i]]);
    exit(0);
  }
  
  /* Initialize */
  for (i=0, ptrb=pflag; i<nop; i++, ptrb++) *ptrb = 0;
  for (i=0, ptri=len, ptrc=ops; i<nop; i++, ptri++, ptrc++) *ptri = (int) strlen(*ptrc); 
  for (i=0, ptri=ncall; i<nop; i++, ptri++) *ptri = 0; 

  for (j=1; j<argc; j++) {
    which = -1;
    opName = argv[j];
    c1 = opName[0];
    for (i=0, ptrc=ops, ptri=len; i<nop; i++, ptrc++, ptri++) {
      if (c1 == **ptrc) {
        temp = strncmp(opName, *ptrc, *ptri);
        if (!temp) {
          which = i; 
          break;
        }
      }
    }

    if (which == -1) {
      print_help(0);
      printf_s("\nERROR: %s is not a valid option\n", argv[j]);
      exit(1);
    }

    pflag[which] = 1;
    pc = argv[j] + len[which];
    count++;
    if (which != 6) opNum = which;
    ncall[which]++;

    switch (which) {
      case 0:
        /* Phenotype file */
        info->pheno_file = pc;
        break;
      case 1:
        /* Genotype file */
        info->geno_file = pc;
        break;
      case 2:
        /* out file */
        info->out_file = pc;
        break;
      case 3:
        /* MAF */
        dval = strtod(pc, &end);
        if (dval < 0.0) {
          printf_s("\n ERROR: max_maf must be positive \n", NULL);
          exit(1);
        }
        info->max_maf = dval;
        break;
      case 4:
        /* N perm */
        buf = strtol(pc, &end, 10);
        info->nperm = (int) buf;
        if (info->nperm < 1) {
          printf_s("\nERROR: nperm must be at least 1\n", NULL);
          exit(1);
        }
        break;
      case 5:
        /* Random seed */
        buf = strtol(pc, &end, 10);
        info->random_seed = (int) buf;
        if (info->random_seed <= 0) {
          printf_s("\nERROR: random_seed must be positive\n", NULL);
          exit(1);
        }
        break;
      case 6:
        /* Help */
        helpFlag = 1;
        break;
      case 7:
        /* Pheno sep */
        info->pheno_sep = get_sep(pc, "pheno_sep");
        break;
      case 8:
        /* Print */
        buf = strtol(pc, &end, 10);
        info->print = (int) buf;
        break;
      case 9:
        /* Gene variable */
        temp = (int) strlen(pc) + 1;
        info->gene_var = (char *) malloc(temp*sizeof(char));
        CHECK_MEM(info->gene_var);
        info->gene_var_len = (int) temp;
        strcpy(info->gene_var, pc);
        break;
      case 10:
        /* CHR size */
        buf = strtol(pc, &end, 10);
        info->chr_size = (int) buf;
        break;
      case 11:
        /* ID size */
        buf = strtol(pc, &end, 10);
        info->id_len = (int) buf;
        if (info->id_len < 1) {
          printf_s("\nERROR: id_len must be positive\n", NULL);
          exit(1);
        }
        info->id_len = info->id_len + 1;
        break;
      case 12:
        /* Pheno buffer */
        buf = strtol(pc, &end, 10);
        if (buf < 5) {
          printf_s("\ERROR: pheno_buffer should be at least 5\n", NULL);
          exit(1);
        }
        info->pheno_buffer = buf;
        break;
      case 13:
        /* Geno buffer */
        buf = strtol(pc, &end, 10);
        if (buf < 20) {
          printf_s("\ERROR: geno_buffer should be at least 20\n", NULL);
          exit(1);
        }
        info->geno_buffer  = buf;
        break;
      case 14:
        /* Geno sep */
        info->geno_sep = get_sep(pc, "geno_sep");
        break;
      case 15:
        /* gene size */
        buf = strtol(pc, &end, 10);
        info->gene_size = (int) buf;
        break;
      case 16:
        /* maxrows */
        buf = strtol(pc, ptrc, 10);
        info->maxrows = buf;
        break;
      case 17:
        /* changesAA variable */
        temp = (int) strlen(pc) + 1;
        info->changesAA_var = (char *) malloc(temp*sizeof(char));
        CHECK_MEM(info->changesAA_var);
        info->changesAA_var_len = (int) temp;
        strcpy(info->changesAA_var, pc);
        break;
      case 18:
        /* functionalClass variable */
        temp = (int) strlen(pc) + 1;
        info->functionalClass_var = (char *) malloc(temp*sizeof(char));
        CHECK_MEM(info->functionalClass_var);
        info->functionalClass_var_len = (int) temp;
        strcpy(info->functionalClass_var, pc);
        break;
      case 19:
        /* gene file */
        info->gene_file = pc;
        info->useGeneTable = 1;
        break;
      case 20:
        /* Version */
        printf_s("%s\n", VERSION);
        exit(0);
        break;
      case 21:
        /* groupedByChr */
        info->groupedByChr = 0;
        break;
      case 22:
        /* Allele missing value (1 character)  */
        info->allele_miss = pc[0];
        break;
      case 23:
        /* maximum chr number */
        buf = (int) strtol(pc, ptrc, 10);
        info->max_chr_number = buf;
        break;
      case 24:
        /* pheno missing value */
        info->pheno_miss = pc[0];
        break;
      case 25:
        /* covar string */
        info->covar_str = pc;
        break;
      case 26:
        /* tests string */
        info->test_str = pc;
        break;
      case 27:
        /* DAS test alpha */
        dval = strtod(pc, &end);
        if ((dval <= 0.0) || (dval >= 1.0)) {
          printf_s("\n ERROR: das_alpha must be between 0 and 1 \n", NULL);
          exit(1);
        }
        info->das_alpha = dval;
        break;
      case 28:
        /* BM test n points */
        temp = (int) strtol(pc, ptrc, 10);
        if ((temp < 1) || (temp > 1000)) {
          printf_s("\n ERROR: bm_n must be between 1 and 1000 \n", NULL);
          exit(1);
        }
        info->bm_n = temp;
        break;
      case 29:
        /* Adapt tolerance */
        dval = strtod(pc, &end);
        if (dval < 0.0) {
          printf_s("\n ERROR: adapt_tol must be positive\n", NULL);
          exit(1);
        }
        info->adapt_tol = dval;
        break;
      case 30:
        /* Adapt maxiter */
        temp = (int) strtol(pc, ptrc, 10);
        if (temp < 1) {
          printf_s("\n ERROR: adapt_maxiter must be >= 1 \n", NULL);
          exit(1);
        }
        info->adapt_maxiter = temp;
        break;
      case 31:
        /* Missing rate */
        dval = strtod(pc, &end);
        if (dval <= 0.0) {
          printf_s("\nERROR: Missing rate must be positive\n", NULL);
          exit(1);
        }
        info->max_missRate = dval;
        break;
      case 32:
        /* Threshold alpha */
        dval = strtod(pc, &end);
        if ((dval <= 0.0) || (dval >= 1.0)) {
          printf_s("\nERROR: Alpha parameter for the Threshold Test must be between 0 and 1\n", NULL);
          exit(1);
        }
        info->threshold_alpha = dval;
        break;
      case 33:
        /* useAllSubsForV */
        info->useControlsForV = 0;
        break;
      case 34:
        /* --no_gene_file */
        info->useGeneTable = 0;
        break;
      case 35:
        /* Rover t */
        dval = strtod(pc, &end);
        if (dval < ALMOST_ZERO) {
          printf_s("\nERROR: rover_t parameter for the Rover Test must positive 1\n", NULL);
          exit(1);
        }
        info->rover_t = dval;
        break;
      case 36:
        /* log file */
        info->log_file = pc;
        break;
      case 37:
        /* include chr list */
        info->chrStr = pc;
        info->inc_chr_flag = 1;
        break;
      case 38:
        /* pheno_header */
        buf = strtol(pc, &end, 10);
        info->pheno_header = (int) buf;
        if ((info->pheno_header != 0) && (info->pheno_header != 1)) {
          printf_s("\nERROR: pheno_header must be set to 0 or 1\n", NULL);
          exit(1);
        }
        break;
      case 39:
        /* response column */
        info->response_var = pc;
        break;
      case 40:
        /* id column */
        info->id_var = pc;
        break;
      case 41:
        /* case value */
        buf = strtol(pc, &end, 10);
        info->case_value = (int) buf;
        break;
      case 42:
        /* control value */
        buf = strtol(pc, &end, 10);
        info->control_value = (int) buf;
        break;
      case 43:
        /* include snps */
        info->snp_file = pc;
        info->inc_snp_flag = 1;
        break;
      case 44:
        /* exclude snps */
        info->snp_file = pc;
        info->exc_snp_flag = 1;
        break;
      case 45:
        /* include genes */
        info->genelist_file = pc;
        info->inc_gene_flag = 1;
        break;
      case 46:
        /* exclude genes */
        info->genelist_file = pc;
        info->exc_gene_flag = 1;
        break;
      case 47:
        /* include subs */
        info->sub_file = pc;
        info->inc_sub_flag = 1;
        break;
      case 48:
        /* exclude subs */
        info->sub_file = pc;
        info->exc_sub_flag = 1;
        break;
      case 49:
        /* exclude chr list */
        info->chrStr = pc;
        info->exc_chr_flag = 1;
        break;
      case 50:
        /* Weight file */
        info->weight_file = pc;
        break;
      case 51:
        /* Weight sep */
        info->weight_sep = get_sep(pc, "weight_sep");
        break;
      case 52:
        /* weight_header */
        buf = strtol(pc, &end, 10);
        info->weight_header = (int) buf;
        if ((info->weight_header != 0) && (info->weight_header != 1)) {
          printf_s("\nERROR: weight_header must be set to 0 or 1\n", NULL);
          exit(1);
        }
        break;
      case 53:
        /* weight_type */
        buf = strtol(pc, &end, 10);
        info->weight_type = (int) buf;
        if ((info->weight_type != 1) && (info->weight_type != 2)) {
          printf_s("\nERROR: weight_type must be set to 1 or 2\n", NULL);
          exit(1);
        }
        break;
      case 54:
        /* INV_ADD */
        dval = strtod(pc, &end);
        if (dval <= 0.0) {
          printf_s("\nERROR: INV_ADD= must be positive\n", NULL);
          exit(1);
        }
        info->INV_ADD = dval;
        break;
      case 55:
        /* INV_DIV */
        dval = strtod(pc, &end);
        if (dval <= 0.0) {
          printf_s("\nERROR: INV_DIV= must be positive\n", NULL);
          exit(1);
        }
        info->INV_DIV = dval;
        break;
      case 56:
        /* Adapt max perm */
        temp = (int) strtol(pc, ptrc, 10);
        if (temp < 0) {
          printf_s("\n ERROR: adapt_maxperm must be >= 0 \n", NULL);
          exit(1);
        }
        info->adapt_maxperm = temp;
        break;
      case 57:
        /* out snp */
        info->out_snp = pc;
        break;
      case 58:
        /* Variance stabilization */
        dval = strtod(pc, &end);
        if (dval < -ALMOST_ZERO) {
          printf_s("\nERROR: vs= must be >= 0\n", NULL);
          exit(1);
        }
        info->vs = dval;
        break;
      case 59:
        /* out permutation file */
        info->out_perm = pc;
        break;
      case 60:
        /* update corr */
        dval = strtod(pc, &end);
        info->update_cor = dval;
        break;
      case 61:
        /* effect size */
        temp = (int) strtol(pc, ptrc, 10);
        if ((temp != 0) && (temp != 1)) {
          printf_s("\n ERROR: effect_size must be 0 or 1 \n", NULL);
          exit(1);
        }
        if (temp) {
          info->effect_size = 1;
        } else {
          info->effect_size = 0;
        }
        break;
      case 62:
        /* minimum MAF */
        dval = strtod(pc, &end);
        if (dval < 0.0) {
          printf_s("\n ERROR: min_maf must be positive \n", NULL);
          exit(1);
        }
        info->min_maf = dval;
        break;
      case 63:
        /* additional columns */
        temp = (int) strtol(pc, ptrc, 10);
        if ((temp != 0) && (temp != 1) && (temp != 2)) {
          printf_s("\n ERROR: addcols must be 0, 1 or 2 \n", NULL);
          exit(1);
        }
        info->add_gcols = temp;
        break;
      case 64:
        /* gz */
        temp = (int) strtol(pc, ptrc, 10);
        if ((temp != 0) && (temp != 1)) {
          printf_s("\n ERROR: geno_gz must be 0 or 1\n", NULL);
          exit(1);
        }
        if (temp) {
          info->gfpt.gzFlag = 1;
        } else {
          info->gfpt.gzFlag = 0;
        }
        break;
      case 65:
        /* sorted by loc */
        temp = (int) strtol(pc, ptrc, 10);
        if ((temp != 0) && (temp != 1)) {
          printf_s("\n ERROR: sortedByLoc must be 0 or 1\n", NULL);
          exit(1);
        }
        if (temp) {
          info->sortedByLoc = 1;
        } else {
          info->sortedByLoc = 0;
        }
        break;
      case 66:
        /* Weight option */
        c1 = tolower(*pc);
        if ((c1 != 'c') && (c1 != 'e') && (c1 != 'n')) {
          printf_s("\n ERROR: weight must be c (codon), e (exon) or n (non-synonomous)\n", NULL);
          exit(1);
        }
        info->weightFlag = 1;
        if (c1 == 'c') {
          info->codonFlag = 1;
        } else if (c1 == 'e') {
          info->exonFlag = 1;
        } else {
          info->nonSynFlag = 1;
        }
        break;
      case 67:
        /* chr alias file */
        info->chr_alias_file = pc;
        break;
      case 68:
        /* store perms */
        temp = (int) strtol(pc, ptrc, 10);
        if ((temp != 0) && (temp != 1)) {
          printf_s("\n ERROR: store_perms must be 0 or 1\n", NULL);
          exit(1);
        }
        info->storePermFlag = temp;
        break;
      case 69:
        /* gene_kb */
        temp = (int) strtol(pc, ptrc, 10);
        if ((temp < 0) || (temp > 1000)) {
          printf_s("\n ERROR: gene_kb must be between 0 and 1000\n", NULL);
          exit(1);
        }
        info->gene_kb = 1000*temp;
        break;
      case 70:
        /* sort_out */
        c1 = tolower(*pc);
        if ((c1 != 'p') && (c1 != 'd')) {
          printf_s("\n ERROR: sort_out must be p or d\n", NULL);
          exit(1);
        }
        if (c1 == 'p') info->sort_out = 1;
        break;
      default:
        print_help(0);
        printf_s("\nERROR: %s is not a valid option\n", pc);
        exit(1);
    } /* END: switch */

  } /* END: for */

  if (helpFlag) {
    /* Print help for only 1 option */
    if (count == 2) {
      print_option(opNum);
    } else {
      print_help(0);
    }
    exit(0);
  }

  /* Check for non-specified options */
  for (i=0; i<2; i++) {
    if (!pflag[req[i]]) {
      printf_s("\nERROR: the option %s must be specified \n", ops[req[i]]);
      flag = 1;
    }
    if (flag) exit(1);
  }

  /* Check ncall vector */
 for (i=0, ptri=ncall; i<nop; i++, ptri++) {
   if (*ptri > 1) {
     printf_s("\nERROR: The option %s is specified multiple times.\n", ops[i]);
     exit(1);
   }
 }

} /* END: parse_options */  

/* Function to return the genotype file type */
static void genoFileType(info)
INFO *info;
{
  char line[10];

  my_fopen(info->geno_file, &info->gfpt);
  my_fgets(line, sizeof(line), &info->gfpt);
  my_fclose(&info->gfpt);
  if (line[0] == '#') {
    info->file_type = VCF;
  } else {
    info->file_type = TPED;
  }

} /* END: genoFileType */

/* Function to get the gene file */
static void getGeneFile(info, argv0)
INFO *info;
char *argv0;  /* Element 0 of argv array. This will give the path to crave */ 
{
  binary use=info->useGeneTable, gf_flag, hg_flag;
  int type=info->file_type, len_cr, len, len_gf=0;
  char *pc, c2, *path;

  if ((!use) && (info->gene_file != NULL)) {
    if (type == VCF) {
      printf_s("\nERROR gene_file= is specified with the option --no_gene_file\n", NULL);
      exit(1);
    } else {
      info->useGeneTable = 1;
      use = 1; 
    }
  }

  /* Gene table not needed for vcf files. Check before */
  if ((type == VCF) && (!use)) return;

  gf_flag = (info->gene_file == NULL) ? 0 : 1; 
  if (gf_flag) len_gf = (int) strlen(info->gene_file);
  len_cr = (int) strlen(argv0);
  len = MAX(len_gf, len_cr) + 30;

  /* See if gene_file is one of hg19 */
  hg_flag = 0;
  if (len_gf == 4) {
    pc = info->gene_file;
    if ((pc[0] == 'h') && (pc[1] == 'g') && (pc[2] == '1')) {
      c2 = pc[3];
      if (c2 == '9') hg_flag = 1;
    }
  }

  path = (char *) malloc(len*sizeof(char)); 
  CHECK_MEM(path);
  if ((hg_flag) || (!gf_flag)) {
    /* Copy crave path to hg_dir */
    getDirName(argv0, len_cr, path); 
  } else {
    /* Copy gene_file to hg_dir */
    strcpy(path, info->gene_file);
  }
  
  /* Add a delimiter if needed */
  len = (int) strlen(path);
  if (path[len-1] != '/') path[len++] = '/';

  if (!gf_flag) c2 = '9';
  if ((hg_flag) || (!gf_flag)) {
    path[len++] = 'h';
    path[len++] = 'g';
    path[len++] = '1';
    path[len++] = c2;
    path[len++] = '/';
    path[len++] = END_OF_STR;
  } 
  info->hg_dir = path;
  info->useGeneTable = 1;

 
} /* END: getGeneFile */

/* Function to get the chromsomes to include */
static void getIncludeChrs(info)
INFO *info;
{
  int i;
  binary *pb;

  if (info->chrStr == NULL) {
    for (i=0, pb=info->includeChr; i<MAX_CHR_NUMBER; i++, pb++) *pb = 1; 
    /*info->includeChr[0] = 0;*/
    info->includeChrMax = info->max_chr_number;
  } else {
    if (info->useGeneTable) {
      info->includeChrMax = getChrsFromStr(info->chrStr, MAX_CHR_NUMBER, info->includeChr, info->chr_alias, info->max_chr_number);
      if (info->exc_chr_flag) {
        pb=info->includeChr;
        for (i=1; i<MAX_CHR_NUMBER; i++) pb[i] = 1 - pb[i]; 
      }
    } else {
      /* Ignore for now */
      if (info->inc_chr_flag) {
        printf_s("***WARNING***: The option include_chr= is ignored\n", NULL);
      } else {
        printf_s("***WARNING***: The option exclude_chr= is ignored\n", NULL);
      }
      for (i=0, pb=info->includeChr; i<MAX_CHR_NUMBER; i++, pb++) *pb = 1; 
      info->chrStr = NULL;
    }
  }

} /* END: getIncludeChrs */

/* Function to read a double value from an open file */
static double read_weight(fpt)
FILE *fpt;
{
  char line[101], *end;
  double ret;

  if (fgets(line, sizeof(line), fpt) != NULL) {
    ret = strtod(line, &end);
  } else {
    ret = 0.0;
  }

  return(ret);

} /* END: read_weight */

/* Function to assign a weight */
double getWeight(info, error)
INFO *info;
int error;
{
  double ret;
  int pos, type=info->weight_type;

  if (!type) {
    return(1.0);
  } else if (type == 1) {
    ret = read_weight(info->weight_fpt);
    if (error) return(0.0);
  } else {
    if (!error) {
      pos = findStrInVec(info->id_str, info->weight_snps, info->weight_n);
      if (pos < 0) {
        ret = 0.0;
      } else {
        ret = info->weight_wgts[pos];
      }
    } else {
      return(0.0);
    }
  }

  if ((ret < NUM_ZERO) && (info->print > PRINT_WEIGHT_FILE)) {
    printf_s("***WARNING*** The weight for SNP %s is 0\n", info->id_str);
  }

  return(ret);

} /* END: getWeight */

/* Function to get weight info */
static void weight_info(info)
INFO *info;
{
  int buffer, nr, nc, i, flag, printFlag, index, len, *id, *pi;
  char sep, line[info->pheno_buffer], *token, *end, **snps, cval[MAX_ID_LEN];
  size_t size;
  FILE *fpt;
  double *wgts, *pd, *tmpw;
  FILEFPT gfpt;

  gfpt.gzFlag = 0;
  buffer = info->pheno_buffer;
  if (info->weight_sep == 0) info->weight_sep = getDelimiter(info->weight_file, buffer, &gfpt);
  sep = info->weight_sep;
  nc = info->weight_type;
  if (!nc) nc = getNcols(info->weight_file, &sep, buffer);
  info->weight_type = nc;
  if (info->weight_header == 2) info->weight_header = getHeader(info->weight_file, sep, nc, buffer);
  if ((!nc) || (nc > 2)) {
    printf_s("\nERROR: The weight file must contain either 1 or 2 columns\n", NULL);
    exit(1);
  }

  size = sizeof(line);
  if (nc == 1) {
    /* Open file and read each weight as each snp is read */
    info->weight_fpt = open_file(info->weight_file, "r");
    if (info->weight_header) {
      fgets(line, size, info->weight_fpt);
      checkEndOfLine(line, 1, buffer, "weight");
    }
    return;
  } else {
    /* Get number of rows to define the vectors */
    nr = getNrows(info->weight_file, buffer);
    printFlag = (info->print > PRINT_WEIGHT_FILE) ? 1 : 0;

    fpt = open_file(info->weight_file, "r");
    if (info->weight_header) {
      fgets(line, size, fpt);
      checkEndOfLine(line, 1, buffer, "weight");
      nr--;
    }
    info->weight_snps = (char **) malloc(nr*sizeof(char *));
    CHECK_MEM(info->weight_snps); 
    tmpw = dVec_alloc(nr, 0, 0.0);
    snps = info->weight_snps;
    wgts = tmpw;
    for (i=0, pd=wgts; i<nr; i++, pd++) *pd = 0.0;

    i = 0;
    while(fgets(line, size, fpt) != NULL) {
      checkEndOfLine(line, 1, buffer, "weight");
      i++;
      if (i > nr) {
        printf_s("\nERROR: Problem reading the weight file\n", NULL);
        exit(1);
      }
      index = i - 1;
      token = line;
      flag = copy_token(cval, token, sep, MAX_ID_LEN);
      if (flag) {
        if (printFlag) printf_i("***WARNING*** Problem on line %d of the weight file\n", i); 
        snps[index] = c0Vec_alloc(1);
        snps[index][0] = END_OF_STR;
        continue;
      } else {
        len = (int) strlen(cval);
        snps[index] = c0Vec_alloc(len+1);
        strcpy(snps[index], cval);
      }
      token = next_token(token, buffer, sep);
      if (token == NULL) {
        if (printFlag) printf_i("***WARNING*** Problem on line %d of the weight file\n", i); 
        continue;
      }
      flag = copy_token(cval, token, sep, MAX_ID_LEN);
      if (flag) {
        if (printFlag) printf_i("***WARNING*** Problem on line %d of the weight file\n", i); 
        continue;
      }  
      wgts[index] = strtod(cval, &end);
    }
    fclose(fpt);
    info->weight_n = nr;

    /* Sort by snp */
    id = iVec_alloc(nr, 0, 0);
    for (i=0, pi=id; i<nr; i++, pi++) *pi = i;
    quicksort_str(0, nr-1, snps, id);
    info->weight_wgts = dVec_alloc(nr, 0, 0.0);
    wgts = info->weight_wgts; 
    for (i=0, pi=id, pd=wgts; i<nr; i++, pi++, pd++) *pd = tmpw[*pi]; 

    free(id);
    free(tmpw);
  }

} /* weight_info */

/* Function to read in the chr aliases */
static void get_chr_alias(info)
INFO *info;
{
  int i, chr, maxchr=info->max_chr_number, flag, len;
  FILEFPT gfpt;
  char line[1001], **alias, *token, *end, cnum[11], sep='\t';
  size_t size;

  if (!info->chr_alias_file) return;

  size = sizeof(line);

  info->chr_alias = (char **) malloc(maxchr*sizeof(char *));
  CHECK_MEM(info->chr_alias);
  alias = info->chr_alias;
  for (i=0; i<maxchr; i++) alias[i] = NULL;

  /* Open the file */
  gfpt.gzFlag = is_gz_file(info->chr_alias_file);
  my_fopen(info->chr_alias_file, &gfpt);
  
  i = 0;
  while(my_fgets(line, size, &gfpt) != NULL) {
    i++;
    checkEndOfLine(line, i, 1000, "chr_alias");
    token = line;
    flag = copy_token(cnum, token, sep, 10);
    if (flag) {
      printf_i("\nERROR: with line %d of the chr alias file\n", i);
      exit(1);
    }
    chr = (int) strtol(token, &end, 10);
    if ((chr < 1) || (chr > maxchr)) {
      printf_i("\nERROR: with line %d of the chr alias file\n", i);
      exit(1);
    }

    token = next_token(token, 1000, sep);
    flag = copy_token(line, token, sep, 10);
    if (flag) {
      printf_i("\nERROR: with line %d of the chr alias file\n", i);
      exit(1);
    }
    len = (int) strlen(line) + 1;
    alias[chr] = (char *) malloc(len*sizeof(char));
    CHECK_MEM(alias[chr]);
    strcpy(alias[chr], line);
  }
  my_fclose(&gfpt);

} /* END: get_chr_alias */

/* Function to initialize the info struct */
void init(info, argv, argc)
INFO *info;
char *argv[];
int argc;
{
  int nr, nc, tempi, i;
  long buffer;
  FILEFPT gfpt;
  char *str;
 
  gfpt.gzFlag = 0;

  /* Global variable */
  FPT_LOG_FILE              = 0;

  info->pheno_buffer        = -1;
  info->geno_buffer         = -1;
  info->geno_sep            = 0;
  info->pheno_sep           = 0;
  info->chr_size            = 0;
  info->n_chr               = 0;
  info->gene_size           = 0;
  info->current_gene_len    = 0;
  info->last_gene_len       = 0;
  info->print               = 0;
  info->gene_var_len        = 0;
  info->maxrows             = 0;
  info->max_maf             = 0.5;
  info->min_maf             = 0.0;
  info->current_pos         = 0;
  info->last_pos            = -1;
  info->random_seed         = -1;
  info->id_len              = -1;
  info->n_case              = 0;
  info->n_control           = 0;
  info->snp_in_mult_genes   = 1;
  info->gene_buffer         = 0;
  info->AF_flag             = 0;
  info->ACAN_flag           = 0;
  info->groupedByChr        = 1;
  info->max_chr_number      = MAX_CHR_NUMBER;
  info->allele_miss         = '0';
  info->useGeneTable        = 1;
  info->pheno_miss          = '.';
  info->n_use_covars        = 0;
  info->n_test              = 0;
  info->das_alpha           = 0.05;
  info->bm_n                = 0;
  info->adapt_tol           = 0.1;
  info->adapt_maxiter       = 1;
  info->adapt_flag          = 0;
  info->errorFlag           = 1;
  info->max_missRate        = 0.8;
  info->threshold_alpha     = 0.05;
  info->useControlsForV     = 1;
  info->geno_cov_flag       = 0;
  info->geno_invcor_flag    = 0;
  info->geno_cor_flag       = 0;
  info->gstar_ncat          = 0;
  info->user_def_flag       = 0;
  info->rover_t             = 0.2;
  info->chrStr              = NULL;
  info->delta2_n            = 0;
  info->pheno_header        = 2;
  info->response_col        = 2;
  info->id_col              = -1;
  info->case_value          = 1;
  info->control_value       = 0;
  info->inc_snp_flag        = 0;
  info->exc_snp_flag        = 0;
  info->out_file            = NULL;
  info->pheno_file          = NULL;
  info->geno_file           = NULL;
  info->log_file            = NULL;
  info->snp_file            = NULL;
  info->snp_vec_n           = 0;
  info->inc_gene_flag       = 0;
  info->exc_gene_flag       = 0;
  info->genelist_file       = NULL;
  info->gene_vec_n          = 0;
  info->inc_sub_flag        = 0;
  info->exc_sub_flag        = 0;
  info->sub_file            = NULL;
  info->response_var        = NULL;
  info->id_var              = NULL;
  info->inc_chr_flag        = 0;
  info->exc_chr_flag        = 0;
  info->chr_vec_n           = 0;
  info->weight_file         = NULL;
  info->weight_fpt          = NULL;
  info->weight_n            = 0;
  info->weight_type         = 0;
  info->weight_sep          = 0;
  info->weight_header       = 2;
  info->store_snp_flag      = 0;
  info->INV_ADD             = 0.1;
  info->INV_DIV             = 1.1;
  info->adapt_maxperm       = 1000000;
  info->nperm               = 500;
  info->current_gene_start  = -9999;
  info->current_gene_end    = -9999;
  info->out_snp             = NULL;
  info->out_fpt_snp         = NULL;
  info->vs                  = 0.005;
  info->update_cor          = 0.3;
  info->effect_size         = 1;
  info->gf_alloc            = 0;
  info->add_gcols           = 0;
  info->gfpt.fpt            = NULL;
  info->gfpt.gzfpt          = NULL;
  info->gfpt.gzFlag         = 2;
  info->sortedByLoc         = 1;
  info->finalChr            = 0;
  info->newChrFlag          = 0;
  info->codonFlag           = 0;
  info->exonFlag            = 0;
  info->nonSynFlag          = 0;
  info->weightFlag          = 0;
  info->chr_alias_file      = NULL;
  info->storePermFlag       = 0;
  info->gene_kb             = 10000;
  info->SKAT_flag           = 0;
  info->CAL_flag            = 0;
  info->sort_out            = 0;

  info->perm_fpt            = NULL;
  info->out_perm            = NULL;

  /* Set all variables that will be allocated to NULL */
  info->use_pheno           = NULL;
  info->CC                  = NULL;
  info->temp_CC             = NULL;
  info->pheno_ids           = NULL;
  info->geno_cols           = NULL;
  info->use_geno            = NULL;
  info->current_chr         = NULL;
  info->last_chr            = NULL;
  info->current_gene        = NULL;
  info->last_gene           = NULL;
  info->genotypes           = NULL;
  info->gene_var            = NULL;
  info->changesAA_var       = NULL;
  info->functionalClass_var = NULL;
  info->gene_file           = NULL;
  info->genes_found         = NULL;
  info->used_chr            = NULL;
  info->chr_alias           = NULL;
  info->covmat              = NULL;
  info->temp_covmat         = NULL;
  info->use_covars          = NULL;
  info->covar_str           = NULL;
  info->Y_norm              = NULL;
  info->geno_cov            = NULL;
  info->geno_cor            = NULL;
  info->geno_invcor         = NULL;
  info->delta               = NULL;
  info->test_str            = NULL;
  info->controls            = NULL;
  info->tests               = NULL;
  info->geno_sdG            = NULL;
  info->geno_varG           = NULL;
  info->gstar_invcor        = NULL;
  info->user_delta          = NULL;
  info->user_cov            = NULL;
  info->user_cor            = NULL;
  info->user_invcor         = NULL;
  info->delta2              = NULL;
  info->geno_invcor2        = NULL;
  info->GENE2               = NULL;
  info->geno_sdG2           = NULL;
  info->snp_vec             = NULL;
  info->id_str              = NULL;
  info->gene_vec            = NULL;
  info->chr_vec             = NULL;
  info->weight_snps         = NULL;
  info->weight_wgts         = NULL;
  info->hg_dir              = NULL;
  info->gene_table          = NULL;
  info->storePermMat        = NULL;
  info->out_save            = NULL;

  /* Parse the options */
  parse_options(info, argv, argc);

  if (info->min_maf > info->max_maf) {
    printf_s("\nERROR: min_maf must be less than max_maf\n", NULL);
    exit(1);
  }

  /* case-control values */
  if (info->case_value == info->control_value) {
    printf_s("\nERROR: case_value must be different than control_value\n", NULL);
    exit(1);
  }

  if (info->id_col == info->response_col) {
    printf_s("\nERROR: id_col must be different than response_col\n", NULL);
    exit(1);
  }

  /* Include/exclude snps, genes, subjects */
  if ((info->inc_snp_flag) && (info->exc_snp_flag)) {
    printf_s("\nERROR: include_snp= and exclude_snp= cannot both be specified\n", NULL);
    exit(1);
  }
  if ((info->inc_gene_flag) && (info->exc_gene_flag)) {
    printf_s("\nERROR: include_gene= and exclude_gene= cannot both be specified\n", NULL);
    exit(1);
  }
  if ((info->inc_sub_flag) && (info->exc_sub_flag)) {
    printf_s("\nERROR: include_sub= and exclude_sub= cannot both be specified\n", NULL);
    exit(1);
  }
  if ((info->inc_chr_flag) && (info->exc_chr_flag)) {
    printf_s("\nERROR: include_chr= and exclude_chr= cannot both be specified\n", NULL);
    exit(1);
  }
  if ((info->weight_file) && (info->weightFlag)) {
    printf_s("\nERROR: weight_file= and weight= cannot both be specified\n", NULL);
    exit(1);
  }

  /* Determine if geno file is .gz */
  if (info->gfpt.gzFlag == 2) {
    info->gfpt.gzFlag = 0;
    str = info->geno_file;
    nr = (int) strlen(str);
    if (nr > 3) {
      if ((str[nr-3] == '.') && (str[nr-2] == 'g') && (str[nr-1] == 'z')) info->gfpt.gzFlag = 1;
    }
  }

  if ((info->gfpt.gzFlag) && (WITH_ZLIB == 0)) {
    printf_s("\nERROR: ZLIB is needed for .gz files. Check the makefile and/or gz= option\n", NULL);
    exit(1);
  }

  if ((info->inc_snp_flag) || (info->exc_snp_flag)) {
    info->id_str = (char *) malloc((MAX_ID_LEN+1)*sizeof(char));
    CHECK_MEM(info->id_str);
  }

  /* Get the chromosomes to include */
  getIncludeChrs(info);
  
  parseTestStr(info);
  tempi = info->adapt_maxperm;
  info->adapt_maxiter = (int) (tempi/info->nperm);
  if (info->adapt_maxiter < 2) {
    info->adapt_flag = 0;
    info->adapt_maxiter = 1;
    if ((tempi) && (info->print)) printf_s("\n***WARNING***: Adaptive permutations will not be run. Increase adapt_maxperm= or decrease nperm=\n", NULL);
  } else {
    info->adapt_flag = 1;
  }
  info->qnorm_alpha = normal01_cdfinv(info->das_alpha);
  info->threshold_phi = normal01_cdfinv(1.0 - info->threshold_alpha);

  /* Get the type of genotype file */
  genoFileType(info);

  /* Get the gene file */
  getGeneFile(info, argv[0]);

  /* A tped must use a gene table */
  if ((info->file_type == TPED) && (!info->useGeneTable)) {
    printf_s("\nERROR: The gene_file= option must be specified when geno_file= is a TPED file\n", NULL);
    exit(1);
  }

  /* Error if a VCF file does not use a gene file and weight= is specified */
  if ((info->file_type == VCF) && (!info->useGeneTable) && (info->weightFlag)) {
    printf_s("\nERROR: The options --no_gene_file and weight= cannot both be specified for a VCF file\n", NULL);
    exit(1);
  }

  if (info->id_len <= 0) info->id_len = MAX_ID_LEN;
  
  /* VCF files should be grouped by chromosome */
  if ((info->file_type == VCF) && (!info->groupedByChr)) {
    /*info->groupedByChr = 1;*/
    printf_s("\nERROR: The option --notGroupedByChr is not valid because a VCF file must be grouped by chromosome\n", NULL);
    exit(1);
  }

  /* Get the buffer size for the phenotype data if needed */
  if (info->pheno_buffer <= 0) {
    buffer = buffer_size(info->pheno_file, 10000, 2, &gfpt);
    if (buffer > 0) info->pheno_buffer = buffer + 1000;
  }

  /* Get the buffer size for the genotype data if needed */
  if (info->geno_buffer <= 0) {
    buffer = buffer_size(info->geno_file, 25000, 3, &info->gfpt);
    if (buffer > 0) {
      if (info->file_type == VCF) {
        info->geno_buffer = buffer + 5000;
      } else {
        info->geno_buffer = buffer + 500;
      }
    }
  }

  /* Get ther buffer size for the gene table */
  /*
  if ((info->useGeneTable) & (info->gene_buffer <= 0)) {
    buffer = buffer_size(info->gene_file, 1000, 10, &gfpt);
    if (buffer > 0) info->gene_buffer = buffer + 100;
  }
  */

  /* Get file delimiters */
  if (info->pheno_sep == 0) info->pheno_sep = getDelimiter(info->pheno_file, buffer, &gfpt);

  if (info->geno_sep == 0) {
    if (info->file_type == VCF) {
      info->geno_sep = '\t';
    } else {
      info->geno_sep = getDelimiter(info->geno_file, info->geno_buffer, &info->gfpt);
    }
  }

  /* For a VCF file check for AF, AC, AN */
  if (info->file_type == VCF) AFACAN_vcf(info);

  /* Get the number of rows and columns in the file */
  buffer = (info->pheno_buffer <= 0) ? 10000 : info->pheno_buffer;
  nr = getNrows(info->pheno_file, buffer);
  nc = getNcols(info->pheno_file, &info->pheno_sep, buffer);
  if (nc < 2) {
    if ((info->file_type == VCF) || ((info->file_type == TPED) && (info->snp_file))) {
      printf_s("\nERROR: pheno_file must contain at least 2 columns.\n", NULL);
      printf_s("ERROR: The option pheno_sep may need to be specified\n", NULL);
      exit(1);
    }
  }
  
  info->pheno_ncol = nc;
  info->ncovs = nc - 2;
  info->nsubs = nr;
  info->pheno_nrow = nr;

  /* Determine if pheno file has a header */
  if (info->pheno_header == 2) info->pheno_header = getHeader(info->pheno_file, info->pheno_sep, nc, buffer);

  /* Set the id column number */
  if (info->id_col == -1) {
    if ((info->file_type == VCF) || (info->inc_sub_flag) || (info->exc_sub_flag)) info->id_col = 1;
  }

  /* Allocate memory for the chr structure */
  alloc_chr_str(info);

  /* Set pheno buffer size if not specified */
  if (info->pheno_buffer <= 0) info->pheno_buffer = 20*nc + 100;
  if (info->geno_buffer <= 0) info->geno_buffer = 2000 + nr*100;

  if (info->chr_size < 2) info->chr_size = MAX_CHR_LEN; 
  if (info->gene_size < 2) info->gene_size = MAX_GENE_LEN; 

  if (!info->useGeneTable) {
    if (info->gene_var_len < 2) {
      info->gene_var = (char *) malloc(17*sizeof(char));
      CHECK_MEM(info->gene_var);
      strcpy(info->gene_var, "reflink.geneName");
    }
    if (info->changesAA_var == NULL) {
      info->changesAA_var = (char *) malloc(17*sizeof(char));
      CHECK_MEM(info->changesAA_var);
      strcpy(info->changesAA_var, "refseq.changesAA");
    }
    info->changesAA_var_len = strlen(info->changesAA_var);
    if (info->functionalClass_var == NULL) {
      info->functionalClass_var = (char *) malloc(23*sizeof(char));
      CHECK_MEM(info->functionalClass_var);
      strcpy(info->functionalClass_var, "refseq.functionalClass");
    }
    info->functionalClass_var_len = strlen(info->functionalClass_var);
  }

  /* Set the random seed */
  if (info->random_seed > 0) {
    srand( (unsigned) info->random_seed );
  } else {
    srand( (unsigned) time(NULL) );
  }

  /* Read file to include/exclude genes. Cannot be done in readGeneFile in case of VCF file */
  if (info->genelist_file != NULL) info->gene_vec = readIds(info->genelist_file, (long) MAX_ID_LEN, 1, 1, &(info->gene_vec_n));

  /* File for weights */
  if (info->weight_file != NULL) weight_info(info);

  /* Set default output file names if not specified */
  if (info->out_file == NULL) {
    info->out_file = (char *) malloc(15*sizeof(char));
    CHECK_MEM(info->out_file);
    strcpy(info->out_file, "./out_gene.txt");
  }
  if (info->out_snp == NULL) {
    info->out_snp = (char *) malloc(14*sizeof(char));
    CHECK_MEM(info->out_snp);
    strcpy(info->out_snp, "./out_snp.txt");
  }

  if (info->groupedByChr == 0) info->sortedByLoc = 0;

  /* Gene table */
  if (info->useGeneTable) {
    info->gene_table = (GENE_TABLE **) malloc(MAX_CHR_NUMBER*sizeof(GENE_TABLE *));
    CHECK_MEM(info->gene_table);
    for (i=0; i<MAX_CHR_NUMBER; i++) info->gene_table[i] = NULL;

    info->genes_found = (int *) malloc(MAX_FOUND*sizeof(int));
    CHECK_MEM(info->genes_found);
  }

  /* Chr alias */
  if (info->chr_alias_file) get_chr_alias(info);


} /* END: init */

/* Function to free the info structure */
void free_info(info)
INFO *info;
{

  free_chr(info, 1);

  if (!info->useGeneTable) {
    if (info->gene_var != NULL) free(info->gene_var);
    if (info->current_gene != NULL) free(info->current_gene);
    if (info->last_gene != NULL) free(info->last_gene);
  }

  if (info->use_geno != NULL) free(info->use_geno);
  if (info->current_chr != NULL) free(info->current_chr);
  if (info->last_chr != NULL) free(info->last_chr);
  if (info->changesAA_var != NULL) free(info->changesAA_var);
  if (info->functionalClass_var != NULL) free(info->functionalClass_var);
  if (info->genotypes != NULL) free(info->genotypes);
  if (info->genes_found != NULL) free(info->genes_found);
  if (info->used_chr != NULL) free(info->used_chr);
  if (info->chr_alias != NULL) matrix_free((void **) info->chr_alias, info->max_chr_number);
  if (info->temp_covmat != NULL) matrix_free((void **) info->temp_covmat, info->nsubs);
  if (info->covmat != NULL) matrix_free((void **) info->covmat, info->nsubs);
  if (info->use_covars != NULL) free(info->use_covars);
  if (info->Y_norm != NULL) free(info->Y_norm);
  if (info->delta != NULL) free(info->delta);
  if (info->controls != NULL) free(info->controls);
  if (info->tests != NULL) free_tests(info);
  if (info->geno_sdG != NULL) free(info->geno_sdG);
  if (info->geno_varG != NULL) free(info->geno_varG);
  if (info->snp_vec != NULL) matrix_free((void **) info->snp_vec, info->snp_vec_n);
  if (info->id_str != NULL) free(info->id_str);
  if (info->gene_vec != NULL) matrix_free((void **) info->gene_vec, info->gene_vec_n);
  if (info->weight_snps != NULL) matrix_free((void **) info->weight_snps, info->weight_n);
  if (info->weight_wgts != NULL) free(info->weight_wgts); 
  if (info->weight_fpt != NULL) fclose(info->weight_fpt);
  if (info->CC) free(info->CC);
  if (info->hg_dir) free(info->hg_dir);
  if (info->storePermMat) matrix_free((void **) info->storePermMat, info->nperm);

  
  if (info->gene_table) {
    gt_free_all(info);
    free(info->gene_table); 
  }

  /* Close files */
  fclose(info->out_fpt);
  fclose(info->out_fpt_snp);
  if (info->perm_fpt) fclose(info->perm_fpt);

} /* END: free_info */

/* Function to parse the covariate string */
static void parseCovStr2(info, vec, vars)
INFO *info;
binary *vec; /* Length nc+1 */
char **vars; /* Unsorted vector of column names */
{
  int len=0, nc=info->pheno_ncol, count=0, i;
  char *str=info->covar_str, **covVars;
  int ycol, idcol, num;

  /* Get the column numbers for response and id */
  if (info->response_var != NULL) info->response_col = getVarNum(info->response_var, vars, nc, 1);
  if (info->id_var != NULL) info->id_col = getVarNum(info->id_var, vars, nc, 1);
    
  ycol  = info->response_col;
  idcol = info->id_col;

  if (str != NULL) len = (int) strlen(str);

  if ((str == NULL) || (len == 0)) {
    /* No covariate string specified */
    info->ncovs = 0;
  } else {
      if ((len == 1) && (str[0] == '0')) {
        /* Do not use covariates */
        info->ncovs = 0;
      } else {
        covVars = strIntoVec(str, ',', &count);
        info->n_use_covars = count;
        for (i=0; i<count; i++) {
          num = getVarNum(covVars[i], vars, nc, 1);
          if (num == ycol) {
            printf_s("\nERROR: with covars=%s\n", str);
            printf_i("Column %d is the column for the outcome.\n", ycol);
            exit(1);
          }
          if ((idcol > -1) && (num == idcol)) {
            printf_s("\nERROR: with covars=%s\n", str);
            printf_i("Column %d is the column for the ids.\n", idcol);
            exit(1);
          }
          if (num < 1) {
            printf_s("\nERROR: with covariate=%s\n", covVars[i]);
            exit(1);
          }
           vec[num] = 1;
        }
        matrix_free((void **) covVars, count);
      }
  }

  /* Set the response and id columns */
  vec[ycol]  = 1;
  if (idcol > -1) vec[idcol] = 1;

} /* END: parseCovStr2 */

/* Function to read the phenotype data */
void getPheno2(info)
INFO *info;
{
  FILE *fpt;
  int i, len, nuse, idlenm1, ncase=0, ncontrol=0, n_use_covars, maxCol, sub_vec_n=0;
  int nc=info->pheno_ncol, j, toklen, col, covFlag, nsubs, nr, add, ycol, idcol, nvars;
  char line[info->pheno_buffer], *token, *end, **ids, sep, flag;
  char cnum[101], miss=info->pheno_miss, **sub_vec=NULL, **vars=NULL;
  size_t size;
  binary *pUse, printFlag, *use_covars, ccFlag=1, idFlag, inc_sub_flag=info->inc_sub_flag;
  long buf=info->pheno_buffer;
  double **covmat, dval, case_value, control_value;
  yType *pCC;

  sep = info->pheno_sep;

  /* Change values depending on if there is a header */
  if (info->pheno_header) {
    info->pheno_nrow--;
    info->nsubs--;

    /* Get the variable names */
    vars = getColNames(info->pheno_file, sep, buf, 0, &nvars);
    if (nvars != nc) {
      printf_s("\nERROR with row 1 of pheno_file\n", NULL);
      exit(1);
    }
  } 

  use_covars = (binary *) malloc((nc+1)*sizeof(binary));
  CHECK_MEM(use_covars);
  bvecinit(use_covars, nc+1, 0);

  /* Determine which covars to use */
  parseCovStr2(info, use_covars, vars);

  /* Check for errors */
  if (info->response_col > nc) {
    printf_i("\nERROR: response column cannot be greater than %d\n", nc);
    exit(1);
  }
  if (info->id_col > nc) {
    printf_i("\nERROR: id column cannot be greater than %d\n", nc);
    exit(1);
  }

  n_use_covars = info->n_use_covars;

  if (vars != NULL) matrix_free((void **) vars, nvars);

  /* Read file to include/exclude subjects */
  if (info->sub_file != NULL) sub_vec = readIds(info->sub_file, (long) MAX_ID_LEN, 1, 1, &sub_vec_n);

  ycol  = info->response_col;
  idcol = info->id_col;

  idFlag = (idcol > -1) ? 1 : 0;

  /* Moved from init */
  nr = info->pheno_nrow;
  info->temp_CC = (yType *) malloc(nr*sizeof(yType));
  CHECK_MEM(info->temp_CC);
  info->use_pheno = (binary *) malloc(nr*sizeof(binary));
  CHECK_MEM(info->use_pheno);
  if (idFlag) {
    info->pheno_ids = (char **) malloc(nr*sizeof(char *));
    CHECK_MEM(info->pheno_ids);
    ids = info->pheno_ids;
    for (i=0; i<nr; i++) ids[i] = NULL;
  }
  bvecinit(info->use_pheno, nr, 0);

  /* Get the maximum column number to use */
  for (i=nc; i>0; i--) {
    if (use_covars[i]) {
      maxCol = i;
      break;
    } 
  }

  size = sizeof(line);
  ids = info->pheno_ids;
  nuse = 0;
  idlenm1 = info->id_len;
  printFlag = (info->print > PRINT_PHENO_FILE) ? 1 : 0;
  nsubs = info->nsubs;
  case_value = (double) info->case_value;
  control_value = (double) info->control_value;

  if (n_use_covars) {
    /* Allocate memory for the covariance matrix */
    info->temp_covmat = (double **) malloc(nsubs*sizeof(double *));
    CHECK_MEM(info->temp_covmat);
    for (i=0; i<nsubs; i++) {
      /* Add 1 column for the intercept */
      info->temp_covmat[i] = (double *) malloc((n_use_covars+1)*sizeof(double));
      CHECK_MEM(info->temp_covmat[i]);
    }
  }
  covmat = info->temp_covmat;
  fpt = open_file(info->pheno_file, "r");
  if (info->pheno_header) {
    fgets(line, size, fpt);
    checkEndOfLine(line, 1, buf, "pheno");
    add = 2;
  } else {
    add = 1;
  }

  /* Read in each line */
  for (i=0, pCC=info->temp_CC, pUse=info->use_pheno; i<nsubs; i++, pCC++, pUse++) {
    fgets(line, size, fpt);
    checkEndOfLine(line, i+add, buf, "pheno");

    /* Initialize */
    token = line;
    covFlag = 0;

    /* Intercept */
    if (n_use_covars) covmat[i][0] = 1.0;
    col = 1;

    for(j=1; j<=maxCol; j++) {
      if (token == NULL) {
        if (printFlag) printf_i("***WARNING***: subject on line %d in pheno_file is being removed\n", i+add);
        covFlag = 1;
        break;
      }

      if (use_covars[j]) {
        /* We need this value */
        flag = copy_token(cnum, token, sep, 99);
        if (flag) {
          if (printFlag) printf_i("***WARNING***: subject on line %d in pheno_file is being removed\n", i+add);
          covFlag = 1;
          break;
        }

        toklen = (int) strlen(cnum);
        if ((toklen == 1) && (cnum[0] == miss)) {
          if (printFlag) printf_i("***WARNING***: subject on line %d in pheno_file is being removed\n", i+add);
          covFlag = 1;
          break;
        }

        /* Detrmine how to store the value */
        if (j == ycol) {
          /* Response */
          dval = strtod(cnum, &end);
          if (ccFlag) {
            if (fabs(dval - control_value) < ALMOST_ZERO) {
              ncontrol++;
              *pCC = 0.0;
            } else if (fabs(dval - case_value) < ALMOST_ZERO) {
              ncase++;
              *pCC = 1.0;
            } else {
              ccFlag = 0;
              *pCC = dval;
            }
          } else {
            *pCC = dval;
          }
        } else if (j == idcol) {
          /* Subject id */
          if (sub_vec_n) {
            /* Check to remove this subject */
            if (removeStr(cnum, sub_vec, sub_vec_n, inc_sub_flag)) {
              covFlag = 1;
              break;
            }
          }
          len = strlen(cnum) + 1;
          ids[i] = (char *) malloc(len*sizeof(char));
          CHECK_MEM(ids[i]);
          strcpy(ids[i], cnum);  
        } else {
          /* A covariate */
          dval = strtod(cnum, &end);
          covmat[i][col] = dval;
          col++;
          if (col > n_use_covars) break;
        }

      } /* END: if (use_covars[j]) */
      token = next_token(token, buf, sep);    
    }
    if (covFlag) continue;  /* Do not use this subject */

    /* Use this ubject */
    *pUse = 1;
    nuse++;
  }
  fclose(fpt);
  free(use_covars);
  if (sub_vec != NULL) matrix_free((void **) sub_vec, sub_vec_n);

  if (nuse == 0) {
    printf_s("\nERROR: No valid rows in pheno_file\n", NULL);
    exit(1);
  }
  info->nsubs = nuse;
  if (ccFlag) {
    info->n_case = ncase;
    info->n_control = ncontrol;
  } else {
    info->n_case = 0;
    info->n_control = 0;
  }
  info->CC_flag = ccFlag;

} /* END: getPheno2 */

/* Function to get the column names and check them */
void getCols(info)
INFO *info;
{
  if (info->file_type == VCF) {
    getCols_vcf(info);
  } else {
    getCols_tped(info);
  }

} /* END: getCols */

/* Function to check the token */
int check_token(token, info)
char *token;
INFO *info;
{
  if (token == NULL) return(1);
  
  return(0);

} /* END: check_token */

/* Function to update the info structure before the genotype data is read */
void update_info(info)
INFO *info;
{
  size_t size;
  int i, ncase=0, ncontrol=0, nsubs=info->nsubs, *ptri;
  yType *pb;
  double sum=0.0, temp;
  binary *pbin;

  if (info->use_pheno != NULL) {
    free(info->use_pheno);
    info->use_pheno = NULL;
  }
  if (info->temp_CC != NULL) {
    free(info->temp_CC);
    info->temp_CC = NULL;
  }
  if (info->pheno_ids != NULL) {
    matrix_free((void **) info->pheno_ids, info->pheno_nrow);
    info->pheno_ids = NULL;
  }
  if (info->geno_cols != NULL) {
    matrix_free((void **) info->geno_cols, info->geno_ncols);
    info->geno_cols = NULL;
  }
  if (info->temp_covmat != NULL) {
    matrix_free((void **) info->temp_covmat, info->pheno_nrow);
    info->temp_covmat = NULL;
  }
  size = (info->chr_size + 1)*sizeof(char);
  info->current_chr = (char *) malloc(size);
  CHECK_MEM(info->current_chr);
  info->last_chr = (char *) malloc(size);
  CHECK_MEM(info->last_chr);

  if (!info->useGeneTable) {
    size = (info->gene_size + 1)*sizeof(char);
    info->current_gene = (char *) malloc(size);
    CHECK_MEM(info->current_gene);
    info->last_gene = (char *) malloc(size);
    CHECK_MEM(info->last_gene);
  }

  if (info->geno_buffer <= 0) info->geno_buffer = (info->n_genotypes)*100 + 2000; 

  /* Compute the number of cases and controls */
  if (info->CC_flag) {
    for (i=0, pb=info->CC; i<nsubs; i++, pb++) {
      temp = *pb;
      if (fabs(temp) < ALMOST_ZERO) {
        ncontrol++;
      } else {
        ncase++;
      }
      sum += temp;
    }
    info->n_case = ncase;
    info->n_control = ncontrol;
    info->test_cc_const = ((double) (ncase*ncontrol*(nsubs - 2.0)))/((double) nsubs);

    if (!ncase) {
      printf_s("\nERROR: no cases in analysis\n", NULL);
      exit(1);
    }
    if (!ncontrol) {
      printf_s("\nERROR: no controls in analysis\n", NULL);
      exit(1);
    }
   
    /* Allocate memory if controls */ 
    info->controls = (binary *) malloc(nsubs*sizeof(binary));
    CHECK_MEM(info->controls);

    for (i=0, pb=info->CC, pbin=info->controls; i<nsubs; i++, pb++, pbin++) {
      if (*pb < 0.5) {
        *pbin = 1;
      } else {
        *pbin = 0;
      }
    }  
  } else {
    info->useControlsForV = 0;
  } 

  info->n2_maf = floor(2.0*(info->n_genotypes)*(info->max_maf));

  info->genotypes = (genoType *) malloc(nsubs*sizeof(genoType));
  CHECK_MEM(info->genotypes);

  /* Vector to check if tped file is grouped by chr */
  if ((info->file_type == TPED) && (info->groupedByChr)) {
    info->used_chr = (int *) malloc(info->max_chr_number*sizeof(int));
    CHECK_MEM(info->used_chr);
    for (i=0, ptri=info->used_chr; i<info->max_chr_number; i++, ptri++) *ptri = 0;
  }

  /* Get the normalized Y */
  getY_norm(info);

  if (info->effect_size == 0) {
    if (info->covmat != NULL) {
      matrix_free((void **) info->covmat, info->nsubs);
      info->covmat = NULL;
    }
    if (info->CC != NULL) {
      free(info->CC);
      info->CC = NULL;
    }
  }

  if (info->gene_vec != NULL) {
    if ((info->file_type != VCF) || ((info->file_type == VCF) && (info->useGeneTable))) {
      matrix_free((void **) info->gene_vec, info->gene_vec_n);
      info->gene_vec = NULL;
      info->gene_vec_n = 0;
    }
  }

  if (info->gf_alloc) {
    free(info->gene_file);
    info->gene_file = NULL;
  }

  /* Matrix of permutations */
  if (info->adapt_maxiter > 1) info->storePermFlag = 0;
  if (info->storePermFlag) getPermMat(info);

} /* END: update_info */

/* Function to read the genotype file */
void read_geno(info)
INFO *info;
{
  char line[info->geno_buffer];
  int *n=&info->geno_line_num, type1Flag, ret, *start=&info->current_gene_start, *end=&info->current_gene_end;
  size_t size;
  long maxrows=info->maxrows, buf=info->geno_buffer;
  FILEFPT *gfpt=&info->gfpt;

  /* Read file to include/exclude snps */
  if (info->snp_file != NULL) info->snp_vec = readIds(info->snp_file, (long) MAX_ID_LEN, 1, 1, &(info->snp_vec_n));

  /* Flag to store the snp id */
  if ((info->snp_vec_n > 0) || (info->weight_type == 2)) info->store_snp_flag = 1;
  if ((info->store_snp_flag) && (info->id_str == NULL)) {
    info->id_str = (char *) malloc((MAX_ID_LEN+1)*sizeof(char));
    CHECK_MEM(info->id_str);
  }

  type1Flag = (info->file_type == VCF);

  size = sizeof(line);
  while (my_fgets(line, size, gfpt) != NULL) {
    *n += 1;
    checkEndOfLine(line, *n, buf, "geno");
    *start = -9999;
    *end   = -9999;
    if (type1Flag) {
      ret = parseLine_vcf(line, info);
    } else {
      ret = parseLine_tped(line, info);
    }
    if ((maxrows) && (*n > maxrows)) break;

  } /* END: while */

} /* END: read_vcf */

/* Function to write out additional columns */
static void write_nperm(info)
INFO *info;
{
  int i, j;
  TEST_STR **tests=info->tests; 

  for (i=0; i<info->n_test; i++) {
    j = tests[i]->which; 
    switch (j) {
      case TEST_SUM:
        fprintf(info->out_fpt, "\t%s", "SUM.Nperm");
        break;
      case TEST_HOTELLING:
        fprintf(info->out_fpt, "\t%s", "HOT.Nperm");
        break;
      case TEST_STOUFFER:
        fprintf(info->out_fpt, "\t%s", "STO.Nperm");
        break;
      case TEST_FISHER:
        fprintf(info->out_fpt, "\t%s", "FI.Nperm");
        break;
      case TEST_CMC:
        fprintf(info->out_fpt, "\t%s", "CMC.Nperm");
        break;
      case TEST_SRT:
        if (info->SKAT_flag) {
          fprintf(info->out_fpt, "\t%s", "SKAT.Nperm");
        } else if (info->CAL_flag) {
          fprintf(info->out_fpt, "\t%s", "CAL.Nperm");
        } else {
          fprintf(info->out_fpt, "\t%s", "MDF.Nperm");
        }
        break;
      case TEST_SRTPOS:
        fprintf(info->out_fpt, "\t%s", "MDP.Nperm");
        break;
      case TEST_DAS:
        fprintf(info->out_fpt, "\t%s", "DAS.Nperm");
        break;
      case TEST_BM:
        fprintf(info->out_fpt, "\t%s", "BM.Nperm");
        break;
      case TEST_STPOSZ:
        fprintf(info->out_fpt, "\t%s", "STP.Nperm");
        break;
      case TEST_THRESHOLD:
        fprintf(info->out_fpt, "\t%s", "TH.Nperm");
        break;
      case TEST_ROVER:
        fprintf(info->out_fpt, "\t%s", "RO.Nperm");
        break;
      case TEST_ROVERV:
        fprintf(info->out_fpt, "\t%s", "RV.Nperm");
        break;
      case TEST_HOTINDEP:
        fprintf(info->out_fpt, "\t%s", "HOI.Nperm");
        break;
      case TEST_C1:
        fprintf(info->out_fpt, "\t%s", "USER_C1.Nperm");
        break;
      case TEST_C2:
        fprintf(info->out_fpt, "\t%s", "USER_C2.Nperm");
        break;
      case TEST_C3:
        fprintf(info->out_fpt, "\t%s", "USER_C3.Nperm");
        break;
      case TEST_C4:
        fprintf(info->out_fpt, "\t%s", "USER_C4.Nperm");
        break;
      case TEST_C5:
        fprintf(info->out_fpt, "\t%s", "USER_C5.Nperm");
        break;
      case TEST_R1:
        fprintf(info->out_fpt, "\t%s", "USER_R1.Nperm");
        break;
      case TEST_R2:
        fprintf(info->out_fpt, "\t%s", "USER_R2.Nperm");
        break;
      case TEST_R3:
        fprintf(info->out_fpt, "\t%s", "USER_R3.Nperm");
        break;
      case TEST_R4:
        fprintf(info->out_fpt, "\t%s", "USER_R4.Nperm");
        break;
      case TEST_R5:
        fprintf(info->out_fpt, "\t%s", "USER_R5.Nperm");
        break;
      default:
        break;
    }
  }

} /* END: write_nperm */

/* Function to write out additional columns */
static void write_ngtobs(info)
INFO *info;
{
  int i, j;
  TEST_STR **tests=info->tests; 

  for (i=0; i<info->n_test; i++) {
    j = tests[i]->which; 
    switch (j) {
      case TEST_SUM:
        fprintf(info->out_fpt, "\t%s", "SUM.N_gt_obs");
        break;
      case TEST_HOTELLING:
        fprintf(info->out_fpt, "\t%s", "HOT.N_gt_obs");
        break;
      case TEST_STOUFFER:
        fprintf(info->out_fpt, "\t%s", "STO.N_gt_obs");
        break;
      case TEST_FISHER:
        fprintf(info->out_fpt, "\t%s", "FI.N_gt_obs");
        break;
      case TEST_CMC:
        fprintf(info->out_fpt, "\t%s", "CMC.N_gt_obs");
        break;
      case TEST_SRT:
        if (info->SKAT_flag) {
          fprintf(info->out_fpt, "\t%s", "SKAT.N_gt_obs");
        } else if (info->CAL_flag) {
          fprintf(info->out_fpt, "\t%s", "CAL.N_gt_obs");
        } else {
          fprintf(info->out_fpt, "\t%s", "MDF.N_gt_obs");
        }
        break;
      case TEST_SRTPOS:
        fprintf(info->out_fpt, "\t%s", "MDP.N_gt_obs");
        break;
      case TEST_DAS:
        fprintf(info->out_fpt, "\t%s", "DAS.N_gt_obs");
        break;
      case TEST_BM:
        fprintf(info->out_fpt, "\t%s", "BM.N_gt_obs");
        break;
      case TEST_STPOSZ:
        fprintf(info->out_fpt, "\t%s", "STP.N_gt_obs");
        break;
      case TEST_THRESHOLD:
        fprintf(info->out_fpt, "\t%s", "TH.N_gt_obs");
        break;
      case TEST_ROVER:
        fprintf(info->out_fpt, "\t%s", "RO.N_gt_obs");
        break;
      case TEST_ROVERV:
        fprintf(info->out_fpt, "\t%s", "RV.N_gt_obs");
        break;
      case TEST_HOTINDEP:
        fprintf(info->out_fpt, "\t%s", "HOI.N_gt_obs");
        break;
      case TEST_C1:
        fprintf(info->out_fpt, "\t%s", "USER_C1.N_gt_obs");
        break;
      case TEST_C2:
        fprintf(info->out_fpt, "\t%s", "USER_C2.N_gt_obs");
        break;
      case TEST_C3:
        fprintf(info->out_fpt, "\t%s", "USER_C3.N_gt_obs");
        break;
      case TEST_C4:
        fprintf(info->out_fpt, "\t%s", "USER_C4.N_gt_obs");
        break;
      case TEST_C5:
        fprintf(info->out_fpt, "\t%s", "USER_C5.N_gt_obs");
        break;
      case TEST_R1:
        fprintf(info->out_fpt, "\t%s", "USER_R1.N_gt_obs");
        break;
      case TEST_R2:
        fprintf(info->out_fpt, "\t%s", "USER_R2.N_gt_obs");
        break;
      case TEST_R3:
        fprintf(info->out_fpt, "\t%s", "USER_R3.N_gt_obs");
        break;
      case TEST_R4:
        fprintf(info->out_fpt, "\t%s", "USER_R4.N_gt_obs");
        break;
      case TEST_R5:
        fprintf(info->out_fpt, "\t%s", "USER_R5.N_gt_obs");
        break;
      default:
        break;
    }
  }

} /* END: write_ngtobs */

/* Function to open the output file */
void open_out(info)
INFO *info;
{
  int i, j;
  TEST_STR **tests=info->tests; 

  info->out_fpt = open_file(info->out_file, "w");

  fprintf(info->out_fpt, "%s\t%s\t%s\t%s\t%s", "Gene", "Chromosome", "Start", "End", "N_SNPs");

  for (i=0; i<info->n_test; i++) {
    j = tests[i]->which; 
    switch (j) {
      case TEST_SUM:
        fprintf(info->out_fpt, "\t%s", "SUM.Pvalue");
        break;
      case TEST_HOTELLING:
        fprintf(info->out_fpt, "\t%s", "HOT.Pvalue");
        break;
      case TEST_STOUFFER:
        fprintf(info->out_fpt, "\t%s", "STO.Pvalue");
        break;
      case TEST_FISHER:
        fprintf(info->out_fpt, "\t%s", "FI.Pvalue");
        break;
      case TEST_CMC:
        fprintf(info->out_fpt, "\t%s", "CMC.Pvalue");
        break;
      case TEST_SRT:
        if (info->SKAT_flag) {
          fprintf(info->out_fpt, "\t%s", "SKAT.Pvalue");
        } else if (info->CAL_flag) {
          fprintf(info->out_fpt, "\t%s", "CAL.Pvalue");
        } else {
          fprintf(info->out_fpt, "\t%s", "MDF.Pvalue");
        }
        break;
      case TEST_SRTPOS:
        fprintf(info->out_fpt, "\t%s", "MDP.Pvalue");
        break;
      case TEST_DAS:
        fprintf(info->out_fpt, "\t%s", "DAS.Pvalue");
        break;
      case TEST_BM:
        fprintf(info->out_fpt, "\t%s", "BM.Pvalue");
        break;
      case TEST_STPOSZ:
        fprintf(info->out_fpt, "\t%s", "STP.Pvalue");
        break;
      case TEST_THRESHOLD:
        fprintf(info->out_fpt, "\t%s", "TH.Pvalue");
        break;
      case TEST_ROVER:
        fprintf(info->out_fpt, "\t%s", "RO.Pvalue");
        break;
      case TEST_ROVERV:
        fprintf(info->out_fpt, "\t%s", "RV.Pvalue");
        break;
      case TEST_HOTINDEP:
        fprintf(info->out_fpt, "\t%s", "HOI.Pvalue");
        break;
      case TEST_C1:
        fprintf(info->out_fpt, "\t%s", "USER_C1.Pvalue");
        break;
      case TEST_C2:
        fprintf(info->out_fpt, "\t%s", "USER_C2.Pvalue");
        break;
      case TEST_C3:
        fprintf(info->out_fpt, "\t%s", "USER_C3.Pvalue");
        break;
      case TEST_C4:
        fprintf(info->out_fpt, "\t%s", "USER_C4.Pvalue");
        break;
      case TEST_C5:
        fprintf(info->out_fpt, "\t%s", "USER_C5.Pvalue");
        break;
      case TEST_R1:
        fprintf(info->out_fpt, "\t%s", "USER_R1.Pvalue");
        break;
      case TEST_R2:
        fprintf(info->out_fpt, "\t%s", "USER_R2.Pvalue");
        break;
      case TEST_R3:
        fprintf(info->out_fpt, "\t%s", "USER_R3.Pvalue");
        break;
      case TEST_R4:
        fprintf(info->out_fpt, "\t%s", "USER_R4.Pvalue");
        break;
      case TEST_R5:
        fprintf(info->out_fpt, "\t%s", "USER_R5.Pvalue");
        break;
      default:
        break;
    }
  }

  if (info->add_gcols > 0) write_nperm(info);
  if (info->add_gcols > 1) write_ngtobs(info);

  fprintf(info->out_fpt, "\n");
  i = fflush(info->out_fpt);

  /* Output file for SNPs */
  info->out_fpt_snp = open_file(info->out_snp, "w");
  fprintf(info->out_fpt_snp, "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s", "Chromosome", "Gene", "SNP_position", "MAF", "NonMiss", "Effect_size", "Delta", "Pvalue");
  fprintf(info->out_fpt_snp, "\n");
  i = fflush(info->out_fpt_snp);

  /* Permutation output */
  if (info->out_perm) {
    info->perm_fpt = open_file(info->out_perm, "w");
    fprintf(info->perm_fpt, "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n", "GENE", "PERM", "BM", "CMC", "DAS", "FI", "HOI", "HOT", "MDF", "MDP", "RO", "RV", "STO", "STP", "SUM", "TH");
    i = fflush(info->perm_fpt);
  }

} /* END: open_out */

/* Function to check for end of line character */
void checkEndOfLine(str, lineNum, n, which)
char *str;
long n;
int lineNum;
char *which;
{
  char c1;
   
  c1 = str[strlen(str)-1];
  if ((c1 != END_OF_LINE) && (c1 != EOF)) {
    printf("\nERROR: End of line character not found on line %d of %s_file\n", lineNum, which);
    printf("%s_buffer was set to %d\n", which, (int) n);
    if (FPT_LOG_FILE) {
      fprintf(FPT_LOG_FILE, "\nERROR: End of line character not found on line %d of %s_file\n", lineNum, which);
      fprintf(FPT_LOG_FILE, "%s_buffer was set to %d\n", which, (int) n);
    }
    exit(1);
  }

} /* END: checkEndOfLine */

/* Function to write results to the output file */
static void writeOut_snp(info, GENE)
INFO *info;
GENE_STR *GENE;
{
  binary group=info->groupedByChr;
  FILE *fpt=info->out_fpt_snp;
  int i, chr_num=info->last_chr_num, nsub=info->nsubs;
  char *chr=info->last_chr, *name=GENE->name;
  SNP_STR *SNP=GENE->first;
  double *pd, pvalue, delta;

  for (i=0, pd=info->delta; i<GENE->n_snps; i++, pd++) {
    delta = *pd;
    pvalue = pvalue_normal01(delta);
    if (group) {
      fprintf(fpt, "%s\t%s\t%d\t%g\t%d\t%g\t%g\t%g", chr, name, SNP->location, SNP->maf, nsub-SNP->nmiss, SNP->effect_size, delta, pvalue);      
    } else {
      fprintf(fpt, "%d\t%s\t%d\t%g\t%d\t%g\t%g\t%g", chr_num, name, SNP->location, SNP->maf, nsub-SNP->nmiss, SNP->effect_size, delta, pvalue);    
    }
    fprintf(fpt, "\n");
    SNP = SNP->next;
  }
  fflush(fpt);

} /* END: writeOut_snp */

/* Function to write results to the output file */
static void writeOutput(info, GENE)
INFO *info;
GENE_STR *GENE;
{
  binary group=info->groupedByChr;
  FILE *fpt=info->out_fpt;
  int i, j;
  TEST_STR **tests=info->tests; 

  if (group) {
    fprintf(fpt, "%s\t%s\t%d\t%d\t%d", GENE->name, info->last_chr, GENE->start, GENE->end, GENE->n_snps);      
  } else {
    fprintf(fpt, "%s\t%d\t%d\t%d\t%d", GENE->name, info->last_chr_num, GENE->start, GENE->end, GENE->n_snps);    
  }
  for (i=0; i<info->n_test; i++) fprintf(fpt, "\t%g", tests[i]->pvalue); 
  if (info->add_gcols) {
    for (i=0; i<info->n_test; i++) fprintf(fpt, "\t%d", tests[i]->nperm);
    if (info->add_gcols > 1) {
      for (i=0; i<info->n_test; i++) {
        if (tests[i]->which != TEST_BM) {
          fprintf(fpt, "\t%d", *(tests[i]->n_gt_obs));
        } else {
          j = min_ivec(tests[i]->n_gt_obs, tests[i]->len);
          fprintf(fpt, "\t%d", j);
        }
      }
    } 
  }
  fprintf(fpt, "\n");
  fflush(fpt);

  writeOut_snp(info, GENE);

} /* END: writeOutput */

/* Function to set the cont vector to determine which tests should be computed */
static int setCont(info)
INFO *info;
{
  int i, n_test=info->n_test, which, ret=0;
  binary V_flag, S_flag, Sinv_flag, Gstar_flag, Sinv2_flag, Sinv12_flag;
  TEST_STR **tests=info->tests, *tst;

  V_flag = (info->geno_cov != NULL) ? 1 : 0;
  S_flag = (info->geno_cor != NULL) ? 1 : 0;
  Sinv_flag = (info->geno_invcor != NULL) ? 1 : 0;
  Sinv2_flag = (info->geno_invcor2 != NULL) ? 1 : 0;
  Gstar_flag = (info->gstar_invcor != NULL) ? 1 : 0;
  Sinv12_flag = (Sinv_flag || Sinv2_flag);

  for (i=0; i<n_test; i++) {
    tst = tests[i];
    tst->cont = 0;
    which = tst->which;

    switch (which) {
        case TEST_SUM:
          if (V_flag) {
            tst->cont = 1;
            ret++;
          }
          break;
        case TEST_HOTELLING:
          if (Sinv12_flag) {
            tst->cont = 1;
            ret++;
          }
          break;
        case TEST_STOUFFER:
          tst->cont = 1;
          ret++;
          break;
        case TEST_FISHER:
          tst->cont = 1;
          ret++;
          break;
        case TEST_CMC:
          if (Gstar_flag) {
            tst->cont = 1;
            ret++;
          }
          break;
        case TEST_SRT:
          if (V_flag) {
            tst->cont = 1;
            ret++;
          }
          break;
        case TEST_SRTPOS:
          if (V_flag) {
            tst->cont = 1;
            ret++;
          }
          break;
        case TEST_DAS:
          if (V_flag) {
            tst->cont = 1;
            ret++;
          }
          break;
        case TEST_BM:
          if (Sinv12_flag) {
            tst->cont = 1;
            ret++;
          }
          break;
        case TEST_STPOSZ:
          tst->cont = 1;
          ret++;
          break;
        case TEST_THRESHOLD:
          if (Sinv12_flag) {
            tst->cont = 1;
            ret++;
          }
          break;
        case TEST_ROVER:
          tst->cont = 1;
          ret++;
          break;
        case TEST_ROVERV:
          if (V_flag) {
            tst->cont = 1;
            ret++;
          }
          break;
        case TEST_HOTINDEP:
          tst->cont = 1;
          ret++;
          break;
        default:
          /* User defined tests */
          if (Sinv_flag) {
            tst->cont = 1;
            ret++;
          }
          break;
    }
  }
  
  return(ret);

} /* END: setCont */
 
/* Function to subtract a small number to each observed test */
static void update_obs_tests(info)
INFO *info;
{
  int i, j;
  TEST_STR **tests=info->tests, *tst;
  double *p;

  for (i=0; i<info->n_test; i++) {
    tst = tests[i]; 
    if (tst->cont) {
      if (!tst->len) {
        *tst->obs -= OBS_FUZZ;
      } else {
        for (j=0, p=tst->obs; j<tst->len; j++, p++) *p -= OBS_FUZZ;
      }
    }
  }

} /* update_obs_tests */

/* Function to set up a gene link */
static void gene_set_vec(GENE, ngeno)
GENE_STR *GENE;
int ngeno;
{
  int nsnp=GENE->n_snps, i, j, n1, n2, nmiss, *pi1, *pi2, *pi3;
  SNP_STR *SNP=GENE->first;
  genoType gg, *pG;

  for (j=0; j<nsnp; j++) {
    n1 = SNP->n1;
    n2 = SNP->n2;
    nmiss = SNP->nmiss; 

    if (n1) {
      SNP->id1 = (int *) malloc(n1*sizeof(int));
      CHECK_MEM(SNP->id1);
    }
    if (n2) {
      SNP->id2 = (int *) malloc(n2*sizeof(int));
      CHECK_MEM(SNP->id2);
    }
    if (nmiss) {
      SNP->idmiss = (int *) malloc(nmiss*sizeof(int));
      CHECK_MEM(SNP->idmiss);
    }

    if ((n1) || (n2) || (nmiss)) {
      pi1 = SNP->id1;
      pi2 = SNP->id2;
      pi3 = SNP->idmiss;
      for (i=0, pG=SNP->G; i<ngeno; i++, pG++) {
        gg = *pG;
        if (!gg) {
          continue;
        } else if (gg == 1) {
          *pi1++ = i;
        } else if (gg == 2) {
          *pi2++ = i;
        } else if (gg == MISS_GENO) {
          *pi3++ = i;
        }
      }
    }

    SNP = SNP->next;
  }

} /* END: gene_set_vec */

/* Function to run permutations and output results */
void runPermsOutput(info)
INFO *info;
{
  CHR_STR *CHR=info->chr[info->chr_str_index];
  int ngenes, gg, nsnp, chr_num=info->last_chr_num, nt, nsubs=info->nsubs, current_pos=info->current_pos;
  GENE_STR *GENE, *PGENE;
  char *chrName=info->last_chr;
  binary group=info->groupedByChr, printFlag, ur1Flag=info->test_flag[TEST_R1], useGeneTab=info->useGeneTable, eff=info->effect_size;
  binary sortedByLoc=info->sortedByLoc, finalChr=info->finalChr, newChrFlag=info->newChrFlag, geneFlag;

  if (info->print > 8) print_chrStr(info);

  ngenes = CHR->n_genes;
  if (!ngenes) return;

  if ((info->print) && (!sortedByLoc)) {
    if (group) {
      printf_s("Run permutations for genes in chromosome %s\n", chrName);
    } else {
      printf_i("Run permutations for genes in chromosome %d\n", chr_num);
    }
  }
  
  printFlag = (info->print > 1);

  GENE = CHR->first;
  geneFlag = 0;

  /* Run permutations for each gene */
  for (gg=0; gg<ngenes; gg++) {

    nsnp = GENE->n_snps;
    if (printFlag) {
      if (group) {
        printf("Start gene %s on chr %s with number of SNPs = %d\n", GENE->name, chrName, nsnp);
        if (FPT_LOG_FILE) fprintf(FPT_LOG_FILE, "Start gene %s on chr %s with number of SNPs = %d\n", GENE->name, chrName, nsnp);
      } else {
        printf("Start gene %s on chr %d with number of SNPs = %d\n", GENE->name, chr_num, nsnp);
        if (FPT_LOG_FILE) fprintf(FPT_LOG_FILE, "Start gene %s on chr %d with number of SNPs = %d\n", GENE->name, chr_num, nsnp);
      }
    }
    if (!nsnp) {
      PGENE = GENE;
      GENE = GENE->next;

      /* Free gene */
      free_gene(PGENE, useGeneTab);
      CHR->n_genes--;
      continue;
    }

    /* Check gene end location if sorted by loc */
    if ((sortedByLoc) && (!finalChr) && (!newChrFlag)) {
      if (GENE->end >= current_pos) {
        geneFlag = 1;
        break;
      }
    }

    /* Define the vectors to 1, 2 and missing genotypes */
    gene_set_vec(GENE, nsubs);

    /* Set up V, sigma, sigma^-1, delta for tests */
    nt = tests_init(info, GENE);

    if (!nt) {

      /* Determine which tests to compute */
      nt = setCont(info);
      if (nt) {
        /* Compute the observed test statistics */
        compute_tests2(info, GENE, 0);
        update_obs_tests(info);

        /* Compute the permutation tests */
        permute_tests(info, GENE);
      }
    }

    /*if (eff) effect_size_all(info, GENE);*/
    if (eff) effect_size(info, GENE);

    /* Output */
    writeOutput(info, GENE);

    /* Free memory for allocated in tests_init */
    if (info->geno_cov != NULL) {
      matrix_free((void **) info->geno_cov, nsnp);
      info->geno_cov = NULL;
    }
    if (info->geno_cor != NULL) {
      matrix_free((void **) info->geno_cor, nsnp);
      info->geno_cor = NULL;
    } 

    if (info->geno_invcor != NULL) {
      matrix_free((void **) info->geno_invcor, nsnp);
      info->geno_invcor = NULL;
    }
    if (info->delta != NULL) {
      free(info->delta);
      info->delta = NULL;
    }

    if (info->geno_sdG != NULL) {
      free(info->geno_sdG);
      info->geno_sdG = NULL;
    }
    if (info->geno_varG != NULL) {
      free(info->geno_varG);
      info->geno_varG = NULL;
    }

    if (info->gstar_invcor != NULL) {
      matrix_free((void **) info->gstar_invcor, info->gstar_ncat);
      info->gstar_invcor = NULL;
    }

    /* For user-defined R test */
    if (ur1Flag) {
      if (info->user_delta != NULL) {
        free(info->user_delta);
        info->user_delta = NULL;
      }
      if (info->user_cov != NULL) {
        free(info->user_cov);
        info->user_cov = NULL;
      }
      if (info->user_cor != NULL) {
        free(info->user_cor);
        info->user_cor = NULL;
      }
      if (info->user_invcor != NULL) {
        free(info->user_invcor);
        info->user_invcor = NULL;
      }
    }

    if (info->delta2 != NULL) {
      free(info->delta2);
      info->delta2 = NULL;
    }
    if (info->geno_invcor2 != NULL) {
      matrix_free((void **) info->geno_invcor2, info->delta2_n);
      info->geno_invcor2 = NULL;
    }
    if (info->geno_sdG2 != NULL) {
      free(info->geno_sdG2);
      info->geno_sdG2 = NULL;
    }
    info->delta2_n = 0;
    if (info->GENE2 != NULL) {
      free_gene(info->GENE2, 1);
      info->GENE2 = NULL;
    }

    PGENE = GENE;
    GENE = GENE->next;

    /* Free gene */
    free_gene(PGENE, useGeneTab);
    CHR->n_genes--;

  }

  if (geneFlag) {
    /* There are still genes for this chr */
    CHR->first = GENE;
  } else {
    info->last_pos = -1;
    CHR->first = NULL;
    CHR->last = NULL;
    CHR->n_genes = 0;

    /* May not be needed to be called anymore */
    free_chr(info, 0);
  }
  if (info->newChrFlag) {
    info->newChrFlag = 0;
    gt_free_all(info);
  }

} /* END: runPermsOutput */

/* Function to call runPermsOutput for each chr */
void chrPermsOutput(info)
INFO *info;
{
  int i;

  info->finalChr = 1;
  if (info->groupedByChr) {
    runPermsOutput(info);
  } else {
    for (i=1; i<info->n_chr_str; i++) {
      info->chr_str_index = i;
      info->last_chr_num = i;
      runPermsOutput(info);
    }
  }

} /* chrPermsOutput */

/* Function to check the position of a snp */
int check_pos(token, info)
char *token;
INFO *info;
{
  int flag;
  char *end, temp[21];
  long pos;

  if (token == NULL) return(1);

  flag = copy_token(temp, token, info->geno_sep, 20);
  if (flag) return(flag);
 
  /* Get the numeric posiiton */
  pos = strtol(temp, &end, 10);

  if (pos <= 0) return(1);

  /* Update the positions */
  info->last_pos = info->current_pos;
  info->current_pos = pos;

  /* If same snp, then discard */
  /*if (pos == info->last_pos) return(1);*/

  return(0);

} /* END: check_pos */

/* Function to return a location */
static locType get_loc(token, ctemp, sep, row)
char *token, *ctemp, sep;
int row;
{
  int flag;
  locType loc;
  char *end;

  flag  = copy_token(ctemp, token, sep, 100);
  loc   = (int) strtol(ctemp, &end, 10);
  if ((flag) || (!loc)) {
    printf_i("\nERROR: with gene start location on line %d of the gene file\n", row);
    exit(1);
  }

  return(loc);

} /* END: get_loc */

/* Function to read a file of non-synonomous SNPs */
static void readNonSynTable(info)
INFO *info;
{
  int nr, chr=info->current_chr_num, row, flag, print, len, i;
  long bufferSize;
  locType loc0, *loc, loc00;
  char *line, ctemp[101], *token, *end, sep='\n';
  size_t size;
  FILEFPT gfpt;
  GENE_TABLE *GT;
  char *chr_num_str[28] = {"0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12", "13", "14", "15",\
                           "16", "17", "18", "19", "20", "21", "22", "23", "24", "25", "26", "27"}; 

  gfpt.gzFlag = 1;
  bufferSize = 1000;
  
  line = (char *) malloc(bufferSize*sizeof(char));
  CHECK_MEM(line);
  size = bufferSize*sizeof(char);

  /* Get the name of the file */
  strcpy(line, info->hg_dir);
  len = (int) strlen(line);
  token = chr_num_str[chr];
  line[len++] = 'c';
  line[len++] = 'h';
  line[len++] = 'r';
  line[len++] = '_';
  line[len++] = token[0];
  if (chr > 9) line[len++] = token[1];
  line[len++] = '.';
  line[len++] = 'n';
  line[len++] = 's';
  line[len++] = '.';
  line[len++] = 'g';
  line[len++] = 'z';
  line[len++] = END_OF_STR;

  if (info->print) printf_s("Reading file %s\n", line);

  /* Open the file */
  my_fopen(line, &gfpt); 

  /* Get the number of rows (includes header)*/
  token = my_fgets(line, size, &gfpt);
  if (!token) {
    printf_s("\nERROR: with line 1 of the gene file\n", NULL);
    exit(1);
  }
  flag = copy_token(ctemp, token, sep, 20);
  if (flag) {
    printf_i("\nERROR: with line %d of the gene file\n", 1);
    exit(1);
  }
  nr = (int) strtol(token, &end, 10);
  if (nr < 3) {
    printf_s("\nERROR: reading non-synonomous table\n", NULL);
    exit(1);
  }
 
  /* Allocate memory. Gene table should already be allocated */
  GT = info->gene_table[chr];
  GT->nonSyn_n   = nr;
  GT->nonSyn_loc = iVec_alloc(nr, 0, 0);

  size  = bufferSize*sizeof(line);
  row   = 1;
  print = info->print;
  loc0  = -1;

  for (i=0, loc=GT->nonSyn_loc; i<nr; i++, loc++) {
    row++;
    my_fgets(line, size, &gfpt);
    checkEndOfLine(line, row, bufferSize, "non-synonomous");
    token = line;
    if (!token) {
      printf_i("\nERROR: Problem with line %d of the non-synonomous file\n", row);
      exit(1);
    }

    /* Gene name */
    flag = copy_token(ctemp, token, sep, 100);
    loc00 = (int) strtol(ctemp, &end, 10);

    /* Check for increasing location within chromosome */
    if (loc0 > loc00) {
      printf_i("\nERROR: on line %d of the non-synonomous file. Location is not sorted by increasing position\n", row);
      exit(1);
    }
    *loc = loc00;
    loc0 = loc00;
  }

  my_fclose(&gfpt);
  free(line);

} /* END: readNonSynTable */

/* Function to read a gene table for 1 chr */
void readGeneTable(info)
INFO *info;
{
  /* Tab-delimited
  307
  Gene	GeneStart	GeneEnd	CodonStart	CodonEnd	ExonCount	ExonStart	ExonEnd
  ROCK1P1	109064	122222	122222	122222	5	109064,116841,117042,118296,120700,	109453,116920,117304,118504,122222,
  */
  int nr, chr=info->current_chr_num, row, index, flag, print, len, gene_kb=info->gene_kb;
  long bufferSize;
  locType *loc_start, *loc_stop, loc0, *ploc;
  locType *codon_start, *codon_stop, **exon_start, **exon_stop;
  char **genes, *line, ctemp[101], *token, sep='\t', *end, sep2=',';
  size_t size;
  int i, *n_exon, nx, j, nxm1;
  FILEFPT gfpt;
  GENE_TABLE *GT;
  binary codonFlag=info->codonFlag, exonFlag=info->exonFlag;
  char *chr_num_str[28] = {"0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12", "13", "14", "15",\
                           "16", "17", "18", "19", "20", "21", "22", "23", "24", "25", "26", "27"}; 

  gfpt.gzFlag = 1;
  bufferSize = MAX_GT_BUFFER;
  
  line = (char *) malloc(bufferSize*sizeof(char));
  CHECK_MEM(line);
  size = bufferSize*sizeof(char);

  /* Get the name of the file */
  strcpy(line, info->hg_dir);
  len = (int) strlen(line);
  token = chr_num_str[chr];
  line[len++] = 'c';
  line[len++] = 'h';
  line[len++] = 'r';
  line[len++] = '_';
  line[len++] = token[0];
  if (chr > 9) line[len++] = token[1];
  line[len++] = '.';
  line[len++] = 't';
  line[len++] = 'x';
  line[len++] = 't';
  line[len++] = '.';
  line[len++] = 'g';
  line[len++] = 'z';
  line[len++] = END_OF_STR;

  if (info->print) printf_s("Reading file %s\n", line);

  /* Open the file */
  my_fopen(line, &gfpt); 

  /* Get the number of rows (includes header)*/
  token = my_fgets(line, size, &gfpt);
  if (!token) {
    printf_s("\nERROR: with line 1 of the gene file\n", NULL);
    exit(1);
  }
  flag = copy_token(ctemp, token, sep, 20);
  if (flag) {
    printf_i("\nERROR: with line %d of the gene file\n", 1);
    exit(1);
  }
  nr = (int) strtol(token, &end, 10) - 1;
  if (nr < 3) {
    printf_s("\nERROR: reading gene table\n", NULL);
    exit(1);
  }
 
  /* Row 2 is a header */
  token = my_fgets(line, size, &gfpt);
  if (!token) {
    printf_s("\nERROR: reading gene table\n", NULL);
    exit(1);
  }

  /* Allocate memory if needed */
  if (!info->gene_table[chr]) info->gene_table[chr] = gt_alloc(nr, info);
  GT = info->gene_table[chr];

  loc_start   = GT->loc_start;
  loc_stop    = GT->loc_stop;
  genes       = GT->gene_name;
  codon_start = GT->codon_start;
  codon_stop  = GT->codon_stop;
  n_exon      = GT->n_exon;
  exon_start  = GT->exon_start;
  exon_stop   = GT->exon_stop;

  size  = bufferSize*sizeof(line);
  row   = 2;
  index = 0;
  print = info->print;
  loc0  = -1;

  for (i=0; i<nr; i++) {
    row++;
    my_fgets(line, size, &gfpt);
    checkEndOfLine(line, row, bufferSize, "gene");
    token = line;
    if (!token) {
      if (print > 2) printf_i("***WARNING***: Problem with line %d of the gene file\n", row);
      exit(1);
    }

    /* Gene name */
    flag = copy_token(ctemp, token, sep, 100);
    len = strlen(ctemp);
    if ((flag) || (len > 20)) {
      if (print > 2) printf_i("***WARNING***: Problem with gene column on line %d of the gene file\n", row);
      exit(1);
    }
    genes[i] = (char *) malloc((len+1)*sizeof(char));
    CHECK_MEM(genes[i]);
    strcpy(genes[i], ctemp);
  
    /* Start location */
    token = next_token(token, bufferSize, sep);
    loc_start[i] = get_loc(token, ctemp, sep, row) - gene_kb;

    /* Check for increasing location within chromosome */
    if (loc0 > loc_start[i]) {
      printf_i("\nERROR: on line %d of the gene file. Gene start location is not sorted by increasing position\n", row);
      exit(1);
    }
    loc0 = loc_start[i];

    /* Stop location */
    token = next_token(token, bufferSize, sep);
    loc_stop[i] = get_loc(token, ctemp, sep, row) + gene_kb;

    if ((!codonFlag) && (!exonFlag)) continue;

    /* Codon start */
    token = next_token(token, bufferSize, sep);
    if (codonFlag) codon_start[i] = get_loc(token, ctemp, sep, row);

    /* Codon stop */
    token = next_token(token, bufferSize, sep);
    if (codonFlag) codon_stop[i] = get_loc(token, ctemp, sep, row);

    if (!exonFlag) continue;

    /* Number of exon start/end positions */
    token     = next_token(token, bufferSize, sep);
    nx        = get_loc(token, ctemp, sep, row);
    n_exon[i] = nx;   
    nxm1      = nx - 1;

    /* Allocate memory */
    exon_start[i] = (locType *) malloc(nx*sizeof(locType));
    CHECK_MEM(exon_start[i]);
    exon_stop[i] = (locType *) malloc(nx*sizeof(locType));
    CHECK_MEM(exon_stop[i]);

    /* Start positions are comma delimited */
    token = next_token(token, bufferSize, sep);
    for (j=0, ploc=exon_start[i]; j<nx; j++, ploc++) {
      *ploc = get_loc(token, ctemp, sep2, row);
      if (j < nxm1) token = next_token(token, bufferSize, sep2);
    }

    /* End positions are comma delimited */
    token = next_token(token, bufferSize, sep);
    for (j=0, ploc=exon_stop[i]; j<nx; j++, ploc++) {
      *ploc = get_loc(token, ctemp, sep2, row);
      if (j < nxm1) token = next_token(token, bufferSize, sep2);
    }

    index++;
  }

  my_fclose(&gfpt);
  free(line);
  
  if (info->nonSynFlag) readNonSynTable(info);

} /* END: readGeneTable */

/* Function to get a gene from a table sorted by chr and start location.
   Return the number of genes found */
int getGeneFromTable(info)
INFO *info;
{
  int i0, i1, mid, mid0, found, *ptrf, i00, i11;
  locType *loc_start, *loc_stop;
  int chr_num=info->current_chr_num, pos=info->current_pos, row, count, max;
  GENE_TABLE *GT;

  if ((chr_num > info->max_chr_number) || (chr_num < 1)) return(1);
  GT = info->gene_table[chr_num];
  if (!GT) {
    readGeneTable(info);
    GT = info->gene_table[chr_num];
  }
  loc_start = GT->loc_start;
  loc_stop  = GT->loc_stop;
  i0        = 0;
  i1        = GT->len - 1;

  if (i1 == -1) return(1);
  if ((pos < loc_start[i0]) || (pos > loc_stop[i1])) return(1);
  i00 = i0;
  i11 = i1;

  max = i1 - i0 + 1;
  count = 0;
  mid0 = -1;
  while (1) {
    mid = (i0 + i1)/2;
    if (mid == mid0) mid = ceil(0.5*(i0 + i1));
    if (pos >= loc_start[mid]) {
      /* Try the right endpoint */
      if (pos <= loc_stop[mid]) {
        row = mid;
        break;
      } else {
        /* Update endpoints */
        i0 = mid;
      }
    } else {
      /* Update endpoints */
      i1 = mid;
    }

    if (mid != mid0) {
      mid0 = mid;
    } else {
      return(1);
    }
    if (count > max) return(1);
    count++;
  }

  /* Find all genes */
  found   = 1;
  ptrf    = info->genes_found; 
  *ptrf++ = row;
  i0      = row - 1; 
  while (found < MAX_FOUND) {
    if ((i0 >= i00) && (pos >= loc_start[i0]) && (pos <= loc_stop[i0])) {
      *ptrf++ = i0;
      found++;
    } else {
      break;
    }
    i0--;
  } 
  i0 = row + 1;
  while (found < MAX_FOUND) {
    if ((i0 <= i11) && (pos >= loc_start[i0]) && (pos <= loc_stop[i0])) {
      *ptrf++ = i0;
      found++;
    } else {
      break;
    }
    i0++;
  } 
 
  /* Copy the gene */
  if (found == 1) {
    info->current_gene       = GT->gene_name[row];
    info->current_gene_start = loc_start[row];
    info->current_gene_end   = loc_stop[row];
  }
  info->n_found = found;

  return(0);

} /* END: getGeneFromTable */

/* Function to determine if the current snp is a codo/exon/non-synonomous snp */
int gt_exclude_snp(info)
INFO *info;
{
  int row=info->genes_found[0], ret=1, i, nx;
  locType pos=info->current_pos, *p1, *p2;
  GENE_TABLE *GT=info->gene_table[info->current_chr_num];

  if (!GT) {
    printf_s("\nInternal coding error\n", NULL);
    exit(0);
  }

  if (info->codonFlag) {
    if ((pos >= GT->codon_start[row]) && (pos <= GT->codon_stop[row])) ret = 0;
  } else if (info->exonFlag) {
    nx = GT->n_exon[row]; 
    for (i=0, p1=GT->exon_start[row], p2=GT->exon_stop[row]; i<nx; i++, p1++, p2++) {
      if ((pos >= *p1) && (pos <= *p2)) {
        ret = 0;
        break;
      }
    }
  } else if (info->nonSynFlag) {
    i = binSearch_int(GT->nonSyn_loc, GT->nonSyn_n, pos);
    if (i >= 0) ret = 0;
  }
  
  return(ret);

} /* gt_exclude_snp */


