/***************************************************************
                                                             
                    CRAVE                  
                                                              
 This file is distributed under the GNU General Public        
 License, Version 2. See the file COPYING.txt for more    
 details                                                      
                                                              
****************************************************************/

#include "./crave.h"


/* Function to get the column names from an open tped file */
void getCols_tped(info)
INFO *info;
{
  char line[info->geno_buffer];
  size_t size;
  int nc, ngeno, i, j, row, n_use_covars;
  long buf=info->geno_buffer;
  binary *pb1, *pb2;
  yType *py1, *py2;
  double **pcov1, **pcov2, *pd1, *pd2;
  
  size = sizeof(line);

  /* Read in the first row */
  if (my_fgets(line, size, &info->gfpt) != NULL) { 
    checkEndOfLine(line, 1, buf, "geno");
  } else {
    printf_s("\nERROR: With row 1 of genotype file\n", NULL);
    exit(1);
  }

  /* Count the number of columns */
  nc = count_char(line, info->geno_sep) + 1;

  /* Number of columns should be 4 + 2*(number of phenotype rows) */
  if (nc != 4 + 2*info->pheno_nrow) {
    printf_s("\nERROR: Number of columns in genotype file and/or number of rows in phenotype file is incorrect\n", NULL);
    exit(1);
  }
  info->geno_ncols = nc;
  ngeno = (nc - 4)/2;
  info->n_genotypes = ngeno;
  info->use_geno = (binary *) malloc(ngeno*sizeof(binary));
  CHECK_MEM(info->use_geno);
 
  /* Determine which genotypes to use from the info->use_pheno */
  pb1 = info->use_geno;
  pb2 = info->use_pheno;
  for (i=0; i<ngeno; i++) pb1[i] = pb2[i]; 
  
  /* Close and re-open the file */
  my_fclose(&info->gfpt);
  my_fopen(info->geno_file, &info->gfpt);
  info->geno_line_num = 0;

  /* nsubs should be the number of subjects being used */
  info->CC = (yType *) malloc(info->nsubs*sizeof(yType));
  CHECK_MEM(info->CC);
  
  /* Copy tmpCC to CC (only subjects being used) */
  py1 = info->CC;
  for (i=0, pb1=info->use_pheno, py2=info->temp_CC; i<info->pheno_nrow; i++, pb1++, py2++) {
    if (*pb1) *py1++ = *py2;
  }

  /* Copy the covariance matrix */
  n_use_covars = info->n_use_covars;
  nc = n_use_covars + 1;
  if (n_use_covars) {
    /* Allocate memory for the covariance matrix */
    info->covmat = dblMatrix_alloc(info->nsubs, nc);

    row = 0;
    pcov1 = info->covmat;
    pcov2 = info->temp_covmat;
    for (i=0, pb1=info->use_pheno; i<info->pheno_nrow; i++, pb1++) {
      if (*pb1) {
        /* Copy row */
        for (j=0, pd1=pcov1[row], pd2=pcov2[i]; j<nc; j++, pd1++, pd2++) *pd1 = *pd2;
        row++; 
      }
    }
  }

} /* END: getCols_tped */ 

/* Function to check the chromsome */
static int check_chr_tped(info)
INFO *info;
{
  char *token=info->current_chr;
  int chr_num;

  /* Get the numeric value */
  chr_num = getChrNum(token, info->max_chr_number, 0, NULL);
  if (chr_num == ERROR_CHR) {
    return(1);
  } else {
    info->current_chr_num = chr_num;
  }

  /* If genotype data is not grouped by chromosome, then return */
  if (!info->groupedByChr) return(0);

  if (!info->n_chr) {
    /* First chr */
    strcpy(info->last_chr, token);
    info->last_chr_num = chr_num;
    info->n_chr = 1;
    info->used_chr[chr_num] = 1;
    return(0);
  }

  /* Compare new chr to the last one */
  if (info->last_chr_num == chr_num) {
    /* Not a new chromosome */
    return(0);
  } else {
    /* Check the grouping */
    if (info->used_chr[chr_num]) {
      printf_s("\nERROR: SNPs are not grouped by chromosome in the genotype file. Set option --notGroupedByChr\n", NULL);
      exit(1);
    } else {
      info->used_chr[chr_num] = 1;
    }

    /* New chromosome */
    info->n_chr++;
    info->newChrFlag = 1;

    /* Run permutations and output */
    runPermsOutput(info);

    strcpy(info->last_chr, token);
    info->last_chr_num = chr_num;
  }

  return(0);

} /* check_chr_tped */

/* Function to return the genotype of a subject in a tped file */
static genoType getGeno_tped(miss, a1, a2, a00, a22, a00Count, a22Count)
char miss, a1, a2, *a00, *a22;
int *a00Count, *a22Count;
{
  binary flag1, flag2, errorFlag;
  char A00, A22;
  
  if (a1 != miss) {
    flag1 = 1;
  } else {
    flag1 = 0;
  }
  if (a2 != miss) {
    flag2 = 1;
  } else {
    flag2 = 0;
  }
  if ((!flag1) && (!flag2)) return(MISS_GENO);

  /* Set a00 (first) and a22 if needed */
  if (!*a00Count) {
    if (flag1) {
      *a00 = a1;
    } else if (flag2) {
      *a00 = a2;
    }
  } 
  if (!*a22Count) {
    if ((flag1) && (*a00 != a1)) {
      *a22 = a1;
    } else if ((flag2) && (*a00 != a2)) {
      *a22 = a2;
    }
  }

  /* Update the allele counts */
  A00 = *a00;
  A22 = *a22;
  errorFlag = 0;

  if (flag1) {
    /* a1 is non-missing */
    if (a1 == A00) {
      *a00Count = *a00Count + 1;
    } else if (a1 == A22) {
      *a22Count = *a22Count + 1;
    } else {
      errorFlag = 1;
    } 
  }
  if (flag2) {
    /* a2 is non-missing */
    if (a2 == A00) {
      *a00Count = *a00Count + 1;
    } else if (a2 == A22) {
      *a22Count = *a22Count + 1;
    } else {
      errorFlag = 1;
    }
  }

  /* Get the genotype */
  if (errorFlag) return(ERROR_GENO);
  if ((flag1) && (flag2)) {
    /* Both alleles are non-missing */
    if (a1 != a2) {
      return(1);
    } else {
      if (a1 == A00) {
        return(0);
      } else {
        return(2);
      }
    }
  } else {
    return(MISS_GENO);
  }

} /* END: getGeno_tped */

/* Function to parse a line from a tped file */
int parseLine_tped(line, info)
char *line;
INFO *info;
{
  /* CHROM	ID xxx  pos genotypes */
  char *token, a1, a2, a00, a22, sep=info->geno_sep, miss;
  binary *p_use, flag00, flag01, flag11;
  int count, n_use, flag, a00Count, a22Count, n2_maf, nmiss, nmiss0, ngeno, mina;
  genoType *geno=info->genotypes, *p_geno, gt;
  long buf=info->geno_buffer;
  double maf;
  GENE_STR *GENE;

  token = line;

  /* Chromosome */
  flag = copy_token(info->current_chr, token, sep, info->chr_size);
  if (flag) {
    getWeight(info, 1);
    if (info->print > PRINT_GENO_FILE) printf_i("***WARNING***: Problem in chromosome field of line %d\n", info->geno_line_num);
    return(flag);
  }
  
  /* Check this chromosome. Valid values are 1-22, 23 (X), 24 (Y), 25 (XY), and 26 (MT) */
  flag = check_chr_tped(info);
  if (flag) {
    getWeight(info, 1);
    if (info->print > PRINT_GENO_FILE) printf_i("***WARNING***: Problem in chromosome field of line %d\n", info->geno_line_num);
    return(flag);
  }

  /* See if we are excluding this chr number */
  if (!info->includeChr[info->current_chr_num]) {
    getWeight(info, 1);
    return(0);
  }

  /* ID */
  token = next_token(token, buf, sep);
  if (info->store_snp_flag) {
    flag = copy_token(info->id_str, token, sep, MAX_ID_LEN);
    if (flag) {
      getWeight(info, 1);
      if (info->print > PRINT_GENO_FILE) printf_i("***WARNING***: Problem in ID field of line %d\n", info->geno_line_num);
      return(flag);
    }
    if (info->snp_vec_n) {
      if (removeStr(info->id_str, info->snp_vec, info->snp_vec_n, info->inc_snp_flag)) {
        getWeight(info, 1);
        return(1);
      }
    }
  }

  /* Get the weight */
  info->current_weight = getWeight(info, 0);

  /* Genetic distance (morgans) */
  token = next_token(token, buf, sep);

  /* POS */
  token = next_token(token, buf, sep);
  if (check_pos(token, info)) {
    if (info->print > PRINT_GENO_FILE) printf_i("***WARNING***: Problem in location field of line %d\n", info->geno_line_num);
    return(1);
  }
  if (info->current_pos == info->last_pos) return(1);
  if (info->sortedByLoc) {
    GENE = info->chr[0]->first;
    if (GENE) {
      /* Check the sort */
      if (info->current_pos < info->last_pos) {
        printf_s("\nERROR: geno file not sorted by location within chromosome. Set sortedByLoc=0.\n", NULL);
        exit(1);
      }
      if (info->current_pos > GENE->end) runPermsOutput(info);
    }
  }

  /* Get the gene name */
  flag = getGeneFromTable(info);
  if (flag) {
    if (info->print > PRINT_GENO_FILE) printf_i("***WARNING***: No gene name for SNP on line %d of the genotype file\n", info->geno_line_num);
    return(flag);
  }
  if ((info->weightFlag) && (gt_exclude_snp(info))) return(1);

  /* Genotypes */
  count  = 0;
  n_use  = 0;
  p_use  = info->use_geno;
  p_geno = geno;
  a00Count = 0;
  a22Count = 0;
  a00 = '0';
  a22 = '0';
  n2_maf = info->n2_maf;
  miss   = info->allele_miss;
  ngeno = info->n_genotypes;
  nmiss0 = floor(info->max_missRate*ngeno);
  nmiss = 0;
  flag00 = flag01 = flag11 = 0;
  for (count=0; count<ngeno; count++) {
    /* Get the 2 alleles */
    token = next_token(token, buf, sep);
    if (token == NULL) {
      if (info->print > PRINT_GENO_FILE) {
        printf("***WARNING***: Problem with genotype field number %d on line %d\n", count+1, info->geno_line_num);
        if (FPT_LOG_FILE) fprintf(FPT_LOG_FILE, "***WARNING***: Problem with genotype field number %d on line %d\n", count+1, info->geno_line_num);
      }
      return(1);
    }
    a1 = toupper(*token);
    token = next_token(token, buf, sep);
    if (token == NULL) {
      if (info->print > PRINT_GENO_FILE) {
        printf("***WARNING***: Problem with genotype field number %d on line %d\n", count+1, info->geno_line_num);
        if (FPT_LOG_FILE) fprintf(FPT_LOG_FILE, "***WARNING***: Problem with genotype field number %d on line %d\n", count+1, info->geno_line_num);
      }
      return(1);
    }
    a2 = toupper(*token);

    /* Get the genotype */
    gt = getGeno_tped(miss, a1, a2, &a00, &a22, &a00Count, &a22Count);
    if (gt == ERROR_GENO) {
      if (info->print > PRINT_GENO_FILE) printf_i("***WARNING***: Problem with genotype fields on line %d\n", info->geno_line_num);
      return(1);
    }

    /* Check if we need to continue processing */
    if ((a00Count > n2_maf) && (a22Count > n2_maf)) return(1);
    if (gt == MISS_GENO) {
      nmiss++;
      if (nmiss > nmiss0) return(1);
    }

    if (*p_use) {
      /* Get the genotype */
      *p_geno = gt;
      n_use++;
      p_geno++;
      if (!gt) {
        flag00 = 1;
      } else if (gt == 1) {
        flag01 = 1;
      } else if (gt == 2) {
        flag11 = 1;
      } 
    }
    p_use++;
  }
  if (n_use != info->nsubs) {
    if (info->print > PRINT_GENO_FILE) printf_i("***WARNING***: Problem with genotype fields on line %d\n", info->geno_line_num);
    return(1);
  }

  /* Get the number of subjects used to compute the MAF */
  count = a00Count + a22Count;
  if (!count) return(1);
  mina = MIN(a00Count, a22Count);
  maf = ((double) mina)/count;
  if ((maf > info->max_maf) || (maf < info->min_maf)) return(1);
 
  /* Make sure not all the used genotypes are the same */
  if (flag00 + flag01 + flag11 < 2) return(1);

  /* Switch genotypes if needed */
  if (a00Count < a22Count) switch_genos(geno, n_use);

  /* If the line was ok, store the genotypes. SNP may be in more than 1 gene */
  add_geno(geno, info, maf);

  return(0);

} /* END: parseLine_tped */
