/***************************************************************
                                                             
                    CRAVE                  
                                                              
 This file is distributed under the GNU General Public        
 License, Version 2. See the file COPYING.txt for more    
 details                                                      
                                                              
****************************************************************/

#include "./crave.h"

/* Function to return the genotype of a subject in a vcf file */
static genoType getGeno_vcf(str, refCount, altCount)
char *str;
int *refCount, *altCount;
{
  char a1, a2;

  /* Genotype strings are of the form: 0/0, 1|0, etc */
  a1 = str[0];
  a2 = str[2];

  if (a1 == '0') {
    *refCount = *refCount + 1;
    if (a2 == '0') {
      *refCount = *refCount + 1;
      return(0);
    } else if (a2 != '.') {
      *altCount = *altCount + 1;
      return(1);
    } else {
      return(MISS_GENO);
    }
  } else if (a1 != '.') {
    *altCount = *altCount + 1;
    if (a2 == '0') {
      *refCount = *refCount + 1;
      return(1);
    } else if (a2 != '.') {
      *altCount = *altCount + 1;
      return(2);
    } else {
      return(MISS_GENO);
    }
  } else {
    if (a2 == '0') {
      *refCount = *refCount + 1;
    } else if (a2 != '.') {
      *altCount = *altCount + 1;
    } 
    return(MISS_GENO);
  }

} /* END: getGeno_vcf */

/* Function to match the pheno and geno ids */
static void match_ids_vcf(info)
INFO *info;
{
  int ngeno, p_nr, p_in_g[info->pheno_nrow], nmatch, i, *pint, j, n_use_covars, k, nc;
  char **gids, **pids;
  binary *p_use, *g_use, p_temp[info->pheno_nrow], *pb;
  yType *pCC, *pCC0, *tmpCC;
  double **tmpCov, **pcov1, **pcov2, *pd1, *pd2;

  ngeno = info->geno_ncols - 9;
  info->n_genotypes = ngeno;
  info->use_geno = (binary *) malloc(ngeno*sizeof(binary));
  CHECK_MEM(info->use_geno);
  g_use = info->use_geno;

  p_nr = info->pheno_nrow;
  pids = info->pheno_ids;
  gids = info->geno_cols + 9;

  /* Copy use_pheno to p_temp */
  for (i=0, p_use=info->use_pheno, pb=p_temp; i<p_nr; i++, p_use++, pb++) *pb = *p_use;
  p_use = info->use_pheno;

  /* Match the ids among the valid phenotype ids */
  nmatch = matchCharVecs(pids, p_temp, p_nr, gids, ngeno, p_use, g_use, p_in_g); 
  if (nmatch < 1) {
    printf_s("\nERROR: No matching ids found in pheno_file and geno_file\n", NULL);
    exit(1);
  }
  info->nsubs = nmatch;

  i = ngeno - nmatch;
  if ((info->print > 2) && (i)) printf_i("***WARNING***: %d subject id(s) not matched in the genotype file \n", i);
  
  /* Get the correct order of the response and (covars) */
  info->CC = (yType *) malloc(nmatch*sizeof(yType));
  CHECK_MEM(info->CC);
  tmpCC = (yType *) malloc(ngeno*sizeof(yType));
  CHECK_MEM(tmpCC);

  n_use_covars = info->n_use_covars;
  nc = n_use_covars + 1;
  if (n_use_covars) {
    /* Allocate memory for the covariance matrix */
    info->covmat = dblMatrix_alloc(nmatch, nc);
    tmpCov = dblMatrix_alloc(ngeno, nc);
  }
  pcov1 = tmpCov;
  pcov2 = info->temp_covmat;

  /* Initialize to a non 0-1 value */
  for (i=0, pCC=tmpCC; i<ngeno; i++, pCC++) *pCC = DOUBLE_MISS;

  /* Now get the CC vector that matches the (matched) subjects in the genotype file in the order of the genotype file */
  for (i=0, p_use=info->use_pheno, pCC0=info->temp_CC, pint=p_in_g; i<p_nr; i++, p_use++, pCC0++, pint++) {
    if (*p_use) {
      tmpCC[*pint] = *pCC0;
      if (n_use_covars) {
        for(j=0, pd1=pcov1[*pint], pd2=pcov2[i]; j<nc; j++, pd1++, pd2++) *pd1 = *pd2;  
      }
    } 
  } 

  pcov1 = info->covmat;
  pcov2 = tmpCov;

  /* Remove the ones that equal DOUBLE_MISS, since these subjects are not being used */
  pCC = info->CC;
  j = 0;
  for (i=0, pCC0=tmpCC; i<ngeno; i++, pCC0++) {
    if(fabs(*pCC0 - DOUBLE_MISS) > ALMOST_ZERO) {
      *pCC++ = *pCC0;
      if (n_use_covars) {
        for (k=0, pd1=pcov1[j], pd2=pcov2[i]; k<nc; k++, pd1++, pd2++) *pd1 = *pd2; 
      }
      j++;
    }
  }

  /* Do an error check */
  if (j != nmatch) {
    printf_s("\nERROR: Internal coding error in getting the matched case-control vector\n", NULL);
    exit(1);
  }

  /* Check the last genotype. There may be an extra tab at the end of each row of the vcf file */
  i = strlen(gids[ngeno-1]);
  if ((!i) && (!g_use[ngeno-1])) {
    if (info->print) printf_s("***WARNING***: Extra delimiter at the end of the genotype file\n", NULL);
    info->n_genotypes = ngeno - 1;
  }

  free(tmpCC);
  if (n_use_covars) matrix_free((void **) tmpCov, ngeno);

} /* match_ids_vcf */

/* Function to check that the required columns are in the vcf file */
void checkCols_vcf(info)
INFO *info;
{
  /*#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT*/
  char *cols[9] = {"#CHROM", "POS", "ID", "REF", "ALT", "QUAL", "FILTER", "INFO", "FORMAT"};
  int i, n;
  char **allCols;
 
  n = info->geno_ncols;
  if (n < 10) {
    printf_s("\n ERROR: Input VCF file is missing columns \n", NULL);
    exit(1);
  }

  allCols = info->geno_cols;
  for (i=0; i<9; i++) {
    if (strcmp(cols[i], allCols[i])) {
      printf("\nERROR: column %d in input VCF file should be %s \n", i+1, cols[i]);
      if (FPT_LOG_FILE) fprintf(FPT_LOG_FILE, "\nERROR: column %d in input VCF file should be %s \n", i+1, cols[i]);
      exit(1);
    }
  }

  match_ids_vcf(info);

} /* END: checkCols_vcf */

/* Function to get the column names from an open vcf file */
void getCols_vcf(info)
INFO *info;
{
  char line[info->geno_buffer];
  size_t size;
  int i, nc, flag=0;
  long buf=info->geno_buffer;
  FILEFPT *gfpt=&info->gfpt; 

  size = sizeof(line);
  i = 1;

  /* Loop till we get to the row with the column names */
  while (my_fgets(line, size, gfpt) != NULL) { 
    checkEndOfLine(line, i, buf, "geno");

    if (!strncmp("#CHROM", line, 6)) {
      flag = 1;
      break;
    }
    i++;
    if (i > 1000000) break;
  }
  if (!flag) {
    printf_s("\nERROR: #CHROM column not found in vcf file\n", NULL);
    exit(1);
  }
  info->geno_line_num = i;

  info->geno_cols  = strIntoVec(line, info->geno_sep, &nc);
  info->geno_ncols = nc;

  /* Check the columns */
  checkCols_vcf(info);

} /* END: getCols_vcf */ 

/* Function to check for AF=, AC=/AN= in a VCF file */
void AFACAN_vcf(info)
INFO *info;
{
  char line[info->geno_buffer], *p;
  size_t size;
  int i, flag=0;
  long buf=info->geno_buffer;
  FILEFPT *gfpt=&info->gfpt;
  
  size = sizeof(line);
  i = 1;

  my_fopen(info->geno_file, gfpt);

  /* Loop till we get to the row with the column names */
  while (my_fgets(line, size, gfpt) != NULL) { 
    checkEndOfLine(line, i, buf, "geno");
    if (flag) break;
    if (!strncmp("#CHROM", line, 6)) flag = 1;
    i++;
    if (i > 10000) break;
  }
  if (!flag) {
    printf_s("\nERROR: #CHROM column not found in vcf file\n", NULL);
    exit(1);
  }
  p = strstr(line, "AF=");
  if (p != NULL) info->AF_flag = 1; 
  flag = 0;
  p = strstr(line, "AC=");
  if (p != NULL) flag = 1;
  if (flag) {
    p = strstr(line, "AN=");
    if (p != NULL) info->ACAN_flag = 1;
  }
  
  my_fclose(gfpt);

} /* END: AFACAN_vcf */ 

/* Function to get the gene name from a vcf file */
static int get_gene_vcf(token, info)
char *token;
INFO *info;
{
  char *p, temp[info->gene_size+1];
  int flag;

  if (info->useGeneTable) {
    flag = getGeneFromTable(info);
    if (flag) {
      if (info->print > PRINT_GENO_FILE) printf_i("***WARNING***: No gene name for SNP on line %d of the genotype file\n", info->geno_line_num);
      return(flag);
    }

    if ((info->weightFlag) && (gt_exclude_snp(info))) return(1);
  } else {
    p = strstr(token, info->gene_var);
    if (p == NULL) {
      if (info->print > PRINT_GENO_FILE) printf_i("***WARNING***: Problem in getting the reflink.geneName in the INFO field of line %d\n", info->geno_line_num);
      return(1);
    }
    flag = getStr2delim(p, info->gene_size, "=", ";", info->geno_sep, temp);
    if (flag) {
      if (info->print > PRINT_GENO_FILE) printf_i("***WARNING***: Problem in getting the reflink.geneName in the INFO field of line %d\n", info->geno_line_num);
      return(flag);
    }

    /* See if we are including this gene */
    if (info->gene_vec_n) {
      if (removeStr(temp, info->gene_vec, info->gene_vec_n, info->inc_gene_flag)) return(1); 
    }

    if (info->snp_in_mult_genes) strcpy(info->last_gene, info->current_gene);
    strcpy(info->current_gene, temp);
  }

  return(0);

} /* END: get_gene_vcf */

/* Function to get the snp weight using the Info column from a vcf file */
static double get_weight_vcf(token, info)
char *token;
INFO *info;
{
  char *p, c1;
  int i, flag;
  double wgt;

  if (info->weight_type) {
    wgt = getWeight(info, 0);
    return(wgt);
  } else if (info->useGeneTable) {
    return(1.0);
  }

  /* Get the value of changesAA (true or false) */ 
  p = strstr(token, info->changesAA_var);
  if (p == NULL) return(0.1);
  p = p + info->changesAA_var_len;
  i = flag = 0;
  while(i < 10) {
    if (*p == '=') {
      p = p + 1;
      flag = 1;
      break;
    } else {
      p = p + 1;
      i = i + 1;
    }
  }
  if (!flag) return(0.1);
  c1 = (char) tolower(*p);
  if (c1 != 't') return(0.1);

  /* Get the value of functionalClass (nonsense, missense, or silent) */
  p = strstr(token, info->functionalClass_var);
  if (p == NULL) return(0.1);
  p = p + info->functionalClass_var_len;
  i = flag = 0;
  while(i < 10) {
    if (*p == '=') {
      p = p + 1;
      flag = 1;
      break;
    } else {
      p = p + 1;
      i = i + 1;
    }
  }
  if (!flag) return(0.1);
  c1 = (char) tolower(*p);
  if (c1 == 'n') {
    return(1.0);
  } else {
    return(0.5);
  }

  return(0.1);

} /* END: get_weight_vcf */

/* Function to check the chromsome */
static int check_chr(info)
INFO *info;
{
  char *token=info->current_chr;
  int chr_num;

  if (!info->n_chr) {
    /* First chr */
    if (info->useGeneTable) {
      chr_num = getChrNum(token, info->max_chr_number, 1, info->chr_alias);
      if (chr_num == ERROR_CHR) {
        if (info->print > PRINT_GENO_FILE) printf_s("***WARNING***: No chromosome number in gene_file for chromosome %s\n", token);
        return(1);
      }  
      info->current_chr_num = chr_num;
      info->last_chr_num    = chr_num;
    } else {
      info->current_chr_num = 0;
      info->last_chr_num    = 0;
    }
    strcpy(info->last_chr, token);
    info->n_chr = 1;
    return(0);
  }

  /* Compare new chr to the last one */
  if (!strcmp(token, info->last_chr)) {
    /* Not a new chromosome */
    info->current_chr_num = info->last_chr_num;
    return(0);
  } else {
    /* New chromosome */
    info->n_chr++;
    info->newChrFlag = 1;

    /* Run permutations and output */
    runPermsOutput(info);

    if (info->useGeneTable) {
      chr_num = getChrNum(token, info->max_chr_number, 1, info->chr_alias);
      if (chr_num == ERROR_CHR) {
        if (info->print > PRINT_GENO_FILE) printf_s("***WARNING***: No chromosome number in gene_file for chromosome %s\n", token);
        return(1);
      }  
      info->current_chr_num = chr_num;
      info->last_chr_num    = chr_num;
    }
    strcpy(info->last_chr, token);
  }

  return(0);

} /* check_chr */

/* Function to check for fields AF or AC and AN in the INFO column */
static double check_AF_AC_AN(token, sep, retFlag, AF_flag, ACAN_flag)
char *token, sep;
int *retFlag; 
binary AF_flag, ACAN_flag;
{
  double maf=-1.0;
  int count, flag, ac, an, veclen=20;
  char *p, temp[veclen+1], *endp;
  
  *retFlag = 1;

  /* First check for AF = allele frequency */
  if (AF_flag) {
    p = strstr(token, "AF=");
    if (p != NULL) { 
      flag = getStr2delim(p, veclen, "=", ";", sep, temp);
      if (!flag) {
        /* AF was there, chack if returned string has more than 1 frequency */
       count = count_char(temp, ',');
        if (!count) {
          /* Only 1 frequency, so convert to a double and return */
          maf = strtod(temp, &endp);
          maf = MIN(maf, 1.0-maf);
          *retFlag = 0;
          return(maf);
        } else {
          /* More than 1 frequency, use the genotypes */
          return(maf);
        }
      }
    }
  } 

  /* AF was not there, so check for both AC and AN */
  if (ACAN_flag) {
    p = strstr(token, "AN=");
    if (p == NULL) return(maf);
    flag = getStr2delim(p, veclen, "=", ";", sep, temp);
    if (!flag) {
      count = count_char(temp, ',');
      if (!count) {
        an = (int) strtol(temp, &endp, 10);
        if (!an) {
          *retFlag = ALL_MISS_GENO;
          return(maf);
        }
        
        /* Now get AC */
        p = strstr(token, "AC=");
        if (p == NULL) return(maf);
        flag = getStr2delim(p, veclen, "=", ";", sep, temp);
        if (!flag) {
          count = count_char(temp, ',');
          if (!count) {
            ac = (int) strtol(temp, &endp, 10);
            maf = ((double) ac)/an;
            maf = MIN(maf, 1.0-maf);
            *retFlag = 0;
            return(maf);
          }
        } /* END: if (!flag) */
      }
    } /* END: if (!flag) */
  }

  return(maf);

} /* END: check_AF_AC_AN */

/* Function to parse a line from a vcf file */
int parseLine_vcf(line, info)
char *line;
INFO *info;
{
  /* #CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT */
  char *token;
  char sep=info->geno_sep;
  binary *p_use, posFlag=0, flag00, flag01, flag11;
  int count, n_use, flag, refCount=0, altCount=0, A_flag, n2_maf, nmiss, nmiss0, ngeno;
  genoType *geno=info->genotypes, *p_geno, gt;
  long buf=info->geno_buffer;
  double maf, maf0, max_maf=info->max_maf;
  GENE_STR *GENE;

  token = line;

  /* Chromosome */
  flag = copy_token(info->current_chr, token, sep, info->chr_size);
  if (flag) {
    getWeight(info, 1);
    if (info->print > PRINT_GENO_FILE) printf_i("***WARNING***: Problem in CHROM field of line %d\n", info->geno_line_num);
    return(flag);
  }

  /* Check this chromosome */
  flag = check_chr(info);
  if (flag) {
    getWeight(info, 1);
    if (info->print > PRINT_GENO_FILE) printf_i("***WARNING***: Problem in CHROM field of line %d\n", info->geno_line_num);
    return(flag);
  }

  /* See if we are excluding this chr number */
  if (!info->includeChr[info->current_chr_num]) {
    getWeight(info, 1);
    return(0);
  }

  /* POS */
  token = next_token(token, buf, sep);
  if (check_pos(token, info)) {
    getWeight(info, 1);
    if (info->print > PRINT_GENO_FILE) printf_i("***WARNING***: Problem in POS field of line %d\n", info->geno_line_num);
    return(1);
  }
  if (info->current_pos == info->last_pos) {
    if (!info->snp_in_mult_genes) {
      getWeight(info, 1);
      return(1);
    }
    posFlag = 1;
  }

  if (info->sortedByLoc) {
    GENE = info->chr[0]->first;
    if (GENE) {
      /* Check the sort */
      if (info->current_pos < info->last_pos) {
        printf_s("\nERROR: geno file not sorted by location within chromosome. Set sortedByLoc=0.\n", NULL);
        exit(1);
      }
      if (info->current_pos > GENE->end) runPermsOutput(info);
    }
  }

  /* ID */
  token = next_token(token, buf, sep);
  if (info->store_snp_flag) {
    flag = copy_token(info->id_str, token, sep, MAX_ID_LEN);
    if (flag) {
      getWeight(info, 1);
      if (info->print > PRINT_GENO_FILE) printf_i("***WARNING***: Problem in ID field of line %d\n", info->geno_line_num);
      return(flag);
    }
    if (info->snp_vec_n) {
      if (removeStr(info->id_str, info->snp_vec, info->snp_vec_n, info->inc_snp_flag)) {
        getWeight(info, 1);
        return(1);
      }
    }
  }

  /* REF */
  token = next_token(token, buf, sep);

  /* ALT */
  token = next_token(token, buf, sep);

  /* QUAL */
  token = next_token(token, buf, sep);

  /* FILTER */
  token = next_token(token, buf, sep);

  /* INFO */
  token = next_token(token, buf, sep);
  if (token == NULL) {
    getWeight(info, 1);
    if (info->print > PRINT_GENO_FILE) printf_i("***WARNING***: Problem in INFO field of line %d\n", info->geno_line_num);
    return(1);
  }

  /* Determine if allele counts or frequencies are given */
  maf0 = check_AF_AC_AN(token, sep, &A_flag, info->AF_flag, info->ACAN_flag);
  if (!A_flag) {
    /* Return code is good, so check the MAF */
    if (maf0 > max_maf) {
      getWeight(info, 1);
      return(1);
    }
  } else if (A_flag == ALL_MISS_GENO) {
    getWeight(info, 1);
    return(A_flag);
  }   

  /* Get the gene name */
  flag = get_gene_vcf(token, info);
  if (flag) {
    getWeight(info, 1);
    return(flag);
  }

  /* If it the same snp as last, check the current and last genes */
  if (posFlag) {
    if (!strcmp(info->current_gene, info->last_gene)) {
      getWeight(info, 1);
      return(1);
    }
  }

  /* Get the weight */
  info->current_weight = get_weight_vcf(token, info);

  /* FORMAT */
  token = next_token(token, buf, sep);
  if (token == NULL) {
    if (info->print > PRINT_GENO_FILE) printf_i("***WARNING***: FORMAT field is NULL on line %d\n", info->geno_line_num);
    return(1);
  }

  /* Genotypes */
  count  = 0;
  n_use  = 0;
  p_use  = info->use_geno;
  p_geno = geno;
  n2_maf = info->n2_maf;
  ngeno = info->n_genotypes;
  nmiss0 = floor(ngeno*info->max_missRate);
  nmiss = 0;
  flag00 = flag01 = flag11 = 0;
  for (count=0; count<info->n_genotypes; count++) {
    token = next_token(token, buf, sep);
    if (token == NULL) {
      if (info->print > PRINT_GENO_FILE) {
        printf("***WARNING***: Problem with genotype field number %d on line %d\n", count+1, info->geno_line_num);
        if (FPT_LOG_FILE) fprintf(FPT_LOG_FILE, "***WARNING***: Problem with genotype field number %d on line %d\n", count+1, info->geno_line_num);
      }
      return(1);
    }

    /* Get the genotype */
    gt = getGeno_vcf(token, &refCount, &altCount);

    /* Check if we need to continue processing */
    if ((refCount > n2_maf) && (altCount > n2_maf)) return(1);
    if (gt == MISS_GENO) {
      nmiss++;
      if (nmiss > nmiss0) return(1);
    }

    if (*p_use) {
      /* Get the genotype */
      *p_geno = gt;
      n_use++;
      p_geno++;

      if (!gt) {
        flag00 = 1;
      } else if (gt == 1) {
        flag01 = 1;
      } else if (gt == 2) {
        flag11 = 1;
      }
    }
    p_use++;
  }
  if (n_use != info->nsubs) {
    if (info->print > PRINT_GENO_FILE) printf_i("***WARNING***: Problem with genotype fields on line %d\n", info->geno_line_num);
    return(1);
  }

  /* Get the number of subjects used to compute the MAF. */
  count = refCount + altCount;
  if (!count) return(1);

  maf = ((double) MIN(refCount, altCount))/count;
  if ((!A_flag) && (fabs(maf - maf0) > 0.01)) {
    if (info->print > PRINT_GENO_FILE) printf_i("***WARNING***: MAF from AF/AN/AC does not match MAF from genotypes on line %d\n", info->geno_line_num);
  }
  if ((maf > max_maf) || (maf < info->min_maf)) return(1);
  
  /* Make sure not all the used genotypes are the same */
  if (flag00 + flag01 + flag11 < 2) return(1);

  /* Switch genotypes if needed */
  if (refCount < altCount) switch_genos(geno, n_use);

  /* If the line was ok, store the genotypes */
  add_geno(geno, info, maf);

  return(0);

} /* END: parseLine_vcf */
